import React, { ReactElement, useCallback, useContext, useEffect, useState } from 'react';
import { Button, Modal } from 'react-bootstrap';
import { Issue, TextFileData } from '../../pseuco-shared-components/fileData';
import TextEditor from '../../pseuco-shared-components/ui/editors/text/TextEditor';
import { CurrentThemeContext } from '../../hybrid-documents-reader-core/ThemeApplicator';
import { InteractiveComponent } from '../../hybrid-documents-reader-core/interactive/InteractiveElement';

import classNames from "classnames";

import "./ProgrammingExercise.scss";
import { UserStudyContext } from '../study/UserStudy';

/**
 * The result of the validation of a full program.
 */
export type ValidationResult<ValidationError> = {
    success: boolean;
} & ({
    success: true;
} | {
    success: false;
    error: ValidationError;
})

export type ValidationErrorVisualizer<ValidationError> = React.FC<{
    validationError: ValidationError,
    data: TextFileData
}>;

export type ProgrammingExerciseConfiguration<ValidationError> = {
    exerciseDescription: ReactElement,
    getWorker: () => Worker,
    template?: string,
    testValidation?: boolean,
    overrideSuccessMessage?: ReactElement,
    validateSyntax: (file: TextFileData) => Issue[],
    ValidationErrorVisualization: ValidationErrorVisualizer<ValidationError>,
    userFacingInputType?: string // "program", "model", etc., as displayed to the user
}

const programmingExerciseBuilder = <ValidationError,>(configuration: ProgrammingExerciseConfiguration<ValidationError>): InteractiveComponent => {
    const ValidationErrorVisualization = configuration.ValidationErrorVisualization;

    const ProgrammingExercise: InteractiveComponent = ({ state, setState, markAsSolved, id }) => {
        const theme = useContext(CurrentThemeContext);
        const userStudy = useContext(UserStudyContext);
    
        const [hideSideBar, setHideSideBar] = useState<boolean>(false);

        const [data, setData] = useState<TextFileData>({
            core: typeof state === "string" ? state : configuration.template ?? "",
            extended: {}
        });
    
        const userFacingInputType = configuration.userFacingInputType ?? "program";

        const [editingInProgress, setEditingInProgress] = useState<boolean>(true); // true if the program is being edited and has not settled yet, i.e. issues may be out of date
    
        const onDataChange = useCallback((newCoreData: string | null) => {
            setData({ core: newCoreData, extended: data.extended });
            setEditingInProgress(true);
            setValidationResult(null);
        }, [data]);
        const onDataSettled = useCallback((coreData: string | null) => {
            setState(coreData);

            if (coreData && coreData.length > 0) {
                const issues = configuration.validateSyntax({ core: coreData, extended: data.extended });
                setData({ core: coreData, extended: { ...data.extended, issues } });
            } else {
                setData({ core: coreData, extended: { ...data.extended, issues: undefined } });
            }
            setEditingInProgress(false);
        }, [data.extended, setState]);
    
        /**
         * The validation worker. Is only kept running during validation, will be terminated (and this state set to `null`) once validation is finished.
         */
        const [validationWorker, setValidationWorker] = useState<Worker | null>(null);

        const [validationWorkerTimeout, setValidationWorkerTimeout] = useState<number | null>(null);
        const [validationWorkerTimeoutExceeded, setValidationWorkerTimeoutExceeded] = useState<boolean>(false);
    
        const [validationResult, setValidationResult] = useState<ValidationResult<ValidationError> | null>(null);
    
        const submitForValidation = () => {
            setValidationResult(null);
            const worker = configuration.getWorker();
            setValidationWorker(worker);
            const timeout = window.setTimeout(() => setValidationWorkerTimeoutExceeded(true), 20000);
            setValidationWorkerTimeout(timeout);
            worker.postMessage({ core: data.core, extended: {} });
            worker.onmessage = ({ data: result }: { data: ValidationResult<ValidationError> }) => {
                setValidationWorker(null);
                clearTimeout(timeout);
                setValidationWorkerTimeout(null);
                setValidationWorkerTimeoutExceeded(false);
                setValidationResult(result);

                if (result.success) {
                    markAsSolved();
                } else {
                    userStudy.reportEvent({ type: "INTERACTIVE_REPORT_ERROR", interactiveType: "interactive", interactiveId: id });
                }
            };
        };
    
        const cancelValidation = () => {
            if (validationWorker !== null) validationWorker.terminate();
            if (validationWorkerTimeout) clearTimeout(validationWorkerTimeout);
            setValidationWorkerTimeout(null);
            setValidationWorkerTimeoutExceeded(false);
            setValidationWorker(null);
        };
    
        useEffect(() => () => {
            if (validationWorker !== null) {
                // either, the validationWorker just finished (then, terminating it is harmless), or the component is unmounting, so we better clean up
                validationWorker.terminate();
                setValidationWorker(null);
            }
        }, [validationWorker]);
    
        const hasErrorFreeProgram = data.core !== null && data.core.length > 0 && (data.extended.issues ?? []).every((issue) => issue.type !== "error");
    
        const [showRestoreTemplateModal, setShowRestoreTemplateModal] = useState<boolean>(false);

        const template = configuration.template;

        const [showTestingNoticeModal, setShowTestingNoticeModal] = useState<boolean>(false);

        return <div className={classNames({ "programming-exercise": true, "hide-side-bar": hideSideBar })}>
            <div className="editor-container">
                <TextEditor data={data} onDataChange={onDataChange} onDataSettled={onDataSettled} type="pseuco" options={{ theme }} readOnly={validationWorker !== null} />
            </div>
            <div className="info-container">
                <Button variant="secondary" className="hide-side-bar" onClick={() => setHideSideBar(true)}>Hide Sidebar</Button>
                { template ? <Button variant="secondary" size="sm" className="restore-template" onClick={() => setShowRestoreTemplateModal(true)}>Restore Template</Button> : null }
                { configuration.exerciseDescription }
                <div className="submit-file-line">
                    <Button variant="success" className="submit-file" onClick={submitForValidation} disabled={!(hasErrorFreeProgram && !editingInProgress && validationWorker == null)}>{ validationWorker !== null ? "Validating…" : "Validate" }</Button>
                    { validationWorker !== null ? <Button variant="danger" className="cancel-validation" onClick={cancelValidation}>Cancel</Button> : null }
                </div>
                { validationWorkerTimeoutExceeded ? <div>
                    <p>Validation is taking longer than usual.</p>
                    <p>This likely means your { userFacingInputType } is significantly more complex than expected. Check for infinite loops and try to simplify your { userFacingInputType } to speed up validation.</p>
                    <p>You can continue waiting or cancel validation.</p>
                </div> : null }
                { validationResult ? <div>
                    { !validationResult.success ? <ValidationErrorVisualization validationError={validationResult.error} data={data} /> : configuration.overrideSuccessMessage ?? <div>
                        <p><strong>All Good ✅</strong></p>
                        <p>Your { userFacingInputType } { configuration.testValidation ? "appears to satisfy" : "satisfies" } the specification.</p>
                    </div> }
                </div> : null }
                { configuration.testValidation ? <>
                    <div className="testing-notice-spacer" />
                    <div className="testing-notice">
                        This exercise uses testing-based validation. <a className="testing-notice-link" onClick={() => setShowTestingNoticeModal(true)}>More Info</a>
                    </div>
                </> : null }
            </div>
            <div className="restore-container">
                <Button variant="secondary" className="restore" onClick={() => setHideSideBar(false)}>Show Sidebar</Button>
            </div>
            <Modal show={showRestoreTemplateModal} onHide={() => setShowRestoreTemplateModal(false)}>
                <Modal.Header closeButton>
                    <Modal.Title>Restore Template?</Modal.Title>
                </Modal.Header>
                <Modal.Body>
                    <p>
                        This will delete your { userFacingInputType } and reset the exercise to its initial state.
                    </p>
                    <p>
                        You should only do this if you want to do the exercise again (or if you messed up and want a fresh start).
                        You cannot undo this action.
                    </p>
                    <p>
                    </p>
                </Modal.Body>
                <Modal.Footer>
                    <Button variant="secondary" type="submit" onClick={() => setShowRestoreTemplateModal(false)} >
                      Cancel
                    </Button>
                    <Button variant="danger" type="submit" onClick={() => {
                        if (!template) throw new Error("Internal inconsistency: Template to restore not found.");
                        setShowRestoreTemplateModal(false);
                        setData({ core: template, extended: {} });
                    }} >
                      Restore Template
                    </Button>
                </Modal.Footer>
            </Modal>
            <Modal show={showTestingNoticeModal} onHide={() => setShowTestingNoticeModal(false)}>
                <Modal.Header closeButton>
                    <Modal.Title>Testing-Based Validation</Modal.Title>
                </Modal.Header>
                <Modal.Body>
                    <p>
                        For performance reasons, in this exercise, your solution isn't <em>verified</em>, but just <em>tested</em>.
                        This means that instead of considering every possible execution of your { userFacingInputType }, only a random subset is checked.
                    </p>
                    <p>
                        This means that not every possible bug can be found during testing.
                        If your solution is rejected, this definitely indicates a problem, but if your solution is accepted, this only means your { userFacingInputType } <em>might</em> be correct.
                    </p>
                </Modal.Body>
                <Modal.Footer>
                    <Button variant="primary" type="submit" onClick={() => setShowTestingNoticeModal(false)} >
                      OK
                    </Button>
                </Modal.Footer>
            </Modal>
        </div>;
    };

    return ProgrammingExercise;
};

export default programmingExerciseBuilder;
