import { ReactElement } from 'react';
import { InteractiveComponent } from '../../../hybrid-documents-reader-core/interactive/InteractiveElement';

import programmingExerciseBuilder, { ProgrammingExerciseConfiguration, ValidationErrorVisualizer } from '../ProgrammingExercise';
import React from 'react';
import { CCSPeggySyntaxError, parseCCS } from '@pseuco/ccs-interpreter';
import { Issue } from '../../../pseuco-shared-components/fileData';
import { CCSSyntaxVariants, CCSValidationError, ccsSyntaxVariants } from './validation';

export type CCSProgrammingConfiguration = {
    exerciseDescription: ReactElement,
    getWorker: () => Worker,
    template?: string,
    testValidation?: boolean,
    overrideSuccessMessage?: ReactElement,
    ccsSyntaxVariant: CCSSyntaxVariants
}

const CCSValidationErrorVisualization: ValidationErrorVisualizer<CCSValidationError> = ({ validationError }) => {
    switch (validationError.type) {
        case "GENERIC": {
            return <div>
                <div>A problem occurred during validation.</div>
                <div>{ validationError.message }</div>
            </div>;
        }
        case "CCS_SEMANTICS_ERROR": {
            return <div>
                <div>The validator was unable to fully compute the semantics of the CCS term you entered.</div>
                <div>{ validationError.message }</div>
            </div>;
        }
        case "STRUCTURE_INCORRECT": {
            return <div>
                <div>The CCS term you entered yields an LTS with incorrect structure.</div>
                <div>{ validationError.message }</div>
            </div>;
        }
        default: {
            throw new Error(`Cannot render validation error of unknown type ${(validationError as CCSValidationError).type}.`);
        }
    }
};

const ccsProgrammingBuilder = (configuration: CCSProgrammingConfiguration): InteractiveComponent => {
    /**
     * Check a CCS file for issues, and return a list of [[Issue]] objects.
     * Should probably live in pseuco-shared-components eventually, but as long as the modern CCS semantics are only used in pseuCo Book, it stays here.
     */
    const validateSyntax: ProgrammingExerciseConfiguration<CCSValidationError>["validateSyntax"] = (file): Issue[] => {
        try {
            parseCCS(file.core ?? "", { features: ccsSyntaxVariants[configuration.ccsSyntaxVariant] });
            return [];
        } catch (e) {
            if (e instanceof CCSPeggySyntaxError) {
                const row = e.location.start.line - 1;
                const column = e.location.start.column - 1;
        
                return [{
                    type: "error",
                    scope: "line",
                    row,
                    column,
                    text: e.name + ": " + e.message + " (in column " + (column + 1) + ")",
                }];
            } else {
                return [{
                    type: "error",
                    scope: "file",
                    text: `An unknown error occurred while checking the CCS term: ${e.message}`
                }];
            }
        }
    };
    
    const CCSProgramming: InteractiveComponent = programmingExerciseBuilder({
        exerciseDescription: configuration.exerciseDescription,
        getWorker: configuration.getWorker,
        template: configuration.template,
        testValidation: configuration.testValidation,
        overrideSuccessMessage: configuration.overrideSuccessMessage,
        validateSyntax,
        ValidationErrorVisualization: CCSValidationErrorVisualization,
        userFacingInputType: "model"
    });

    return CCSProgramming;
};

export default ccsProgrammingBuilder;
