import { CCSProcess, CCSSemanticContext, generateSatisfiedRuleApplications } from "@pseuco/ccs-interpreter";
import { LTS } from "../../../pseuco-shared-components/lts/lts";
import { CCSValidator } from "./validation";

export type IsomorphismValidatorConfiguration = {
    lts: LTS
};

/**
 * Confirms a CCS term is isomorphic to a provided LTS.
 * Only works if the sample LTS is deterministic.
 */
export const deterministicIsomorphismValidator = (configuration: IsomorphismValidatorConfiguration): CCSValidator => (ccs) => {
    const lts = configuration.lts;
    const bijection: Map<string, CCSProcess> = new Map(); // maps state from LTS to states from the passed CCS term's semantics

    const semanticContext = new CCSSemanticContext(ccs.environment, {});

    /**
     * Validates that a specific state matches a specific CCS process.
     * Assumes that elements of the bijection are valid or will be checked by someone else. Adds newly-discovered elements to the bijection 
     */
    const checkState = (ltsStateKey: string, ccs: CCSProcess): ReturnType<CCSValidator> => {
        const ltsState = lts.states[ltsStateKey];
        if (!ltsState.transitions) return { success: false, error: { type: "GENERIC", message: `The LTS to compare against was missing transitions in state ${ltsStateKey}.` } };

        const newPairs: [string, CCSProcess][] = [];

        const ruleSolverResult = generateSatisfiedRuleApplications(ccs, semanticContext);
        if (ruleSolverResult.errors.length) {
            return { success: false, error: { type: "CCS_SEMANTICS_ERROR", message: `The CCS term cannot be evaluated: ${ruleSolverResult.errors[0].message}` } };
        }
        const ccsSteps = ruleSolverResult.derivations.map((d) => d.getStep());

        for (const ltsTransition of ltsState.transitions) {
            const matchingCCSSteps = ccsSteps.filter((s) => s.action.toString() === ltsTransition.label);
            
            if (matchingCCSSteps.length < 1) return { success: false, error: { type: "STRUCTURE_INCORRECT", message: `The CCS term ${ccs.toString()} was expected to have a transition with action ${ltsTransition.label}, but none was found.` } };
            if (matchingCCSSteps.length > 1) return { success: false, error: { type: "STRUCTURE_INCORRECT", message: `The CCS term ${ccs.toString()} was expected to have only one transition with action ${ltsTransition.label}, but ${matchingCCSSteps.length} were found.` } }; // only works because we assume the LTS to be deterministic (otherwise, we'd have to try all possible matchings)

            const matchingCCSStep = matchingCCSSteps[0];
            bijection.set(ltsTransition.target, matchingCCSStep.target);
            newPairs.push([ltsTransition.target, matchingCCSStep.target]);
        }

        // check all newly-generated pairs, raise all relevant errors
        for (const [ltsStateKey, ccs] of newPairs) {
            const res = checkState(ltsStateKey, ccs);
            if (!res.success) return res;
        }

        return { success: true };
    };

    return checkState(lts.initialState, ccs.system);
};
