import { TextFileData } from "../../../pseuco-shared-components/fileData";
import { ValidationResult } from "../ProgrammingExercise";
import { CCS, CCSParseOptions, parseCCS } from "@pseuco/ccs-interpreter";

export type CCSValidationError = {
    type: "GENERIC",
    message: string
} | {
    type: "CCS_SEMANTICS_ERROR",
    message: string
} | {
    type: "STRUCTURE_INCORRECT", // the validator checked the structure of the resulting LTS and found a problem
    message: string
};

export type CCSSyntaxVariants = "ccs0" | "ccs0omega";

export const ccsSyntaxVariants: { [variant in CCSSyntaxVariants]: CCSParseOptions["features"] } = {
    "ccs0": {
        parallel: false,
        choice: true,
        sequence: false,
        exit: false,
        restriction: false,
        condition: false,
        processApplication: false,
        valuePassing: false,
        processDefinitions: false
    },
    "ccs0omega": {
        parallel: false,
        choice: true,
        sequence: false,
        exit: false,
        restriction: false,
        condition: false,
        processApplication: true,
        valuePassing: false,
        processDefinitions: true
    }
};

export type CCSValidator = (ccs: CCS) => ValidationResult<CCSValidationError>;

/**
 * Expected to be called in a WebWorker. Registers the `onmessage` callback to receive programs and run the validator against it.
 * @param validator The [[CCSValidator]] to use in this worker.
 */
export const registerValidationWorkerCallback = (validator: CCSValidator, ccsSyntaxVariant: CCSSyntaxVariants): void => {
    const validateProgram = (file: TextFileData): ValidationResult<CCSValidationError> => {
        try {
            let ccs: CCS;
            try {
                ccs = parseCCS(file.core ?? "", { features: ccsSyntaxVariants[ccsSyntaxVariant] });
            } catch (e) {
                return {
                    success: false,
                    error: {
                        type: "GENERIC",
                        message: "Validation failed because the CCS term could not be parsed."
                    }
                };
            }
    
            return validator(ccs);
        } catch (e) {
            return {
                success: false,
                error: {
                    type: "GENERIC",
                    message: `An internal error occurred during validation: ${e.message}`
                }
            };
        }
    };

    self.onmessage = ({ data: file }: { data: TextFileData }) => {
        self.postMessage(validateProgram(file));
    };
};
