import { parseCCS, CCSSemanticContext, CCSUnfulfilledPremise } from "@pseuco/ccs-interpreter";
import React, { useMemo, useState, useEffect } from "react";
import { Container } from "react-bootstrap";
import { InteractiveComponent } from "../../../hybrid-documents-reader-core/interactive/InteractiveElement";
import { ProofBuildingStep, restoreInteractiveCCSProofTreeState, objectifyInteractiveProofTreeState, InteractiveCCSProofTree, CCSRuleSet } from "../../../pseuco-shared-components/ui/ccsDerivation/InteractiveProofTree";
import MathSpan from "../../../pseuco-shared-components/ui/mathSpan/MathSpan";
import { RuleApplicationProofBuildingAction } from "../../../pseuco-shared-components/ui/ccsDerivation/proofBuildingActions";

const makeCCSProofTreeExercise = ({
    ccs,
    enabledRuleSet,
    solutionMode,
    customInstruction
}: {
    ccs: string
    enabledRuleSet?: CCSRuleSet,
    solutionMode: {
        type: "finish-the-tree"
    } | {
        type: "find-enough-transitions",
        minTransitionsToFind: number
    } | {
        type: "open-the-rule-picker"
    } | {
        type: "reach-number-of-steps",
        onlyCountRuleApplications?: string[],
        stepCount: number
    } | {
        type: "attempt-to-reveal-failed-rule-application"
    },
    customInstruction?: string
}) => {
    const CCSProofTreeExercise: InteractiveComponent = ({ state, setState, markAsSolved }) => {
        const ast = useMemo(() => parseCCS(ccs), []);
        const context = useMemo(() => new CCSSemanticContext(ast.environment), [ast.environment]);
        const initialTree = useMemo(() => new CCSUnfulfilledPremise(ast.system), [ast.system]);
    
        const [steps, setSteps] = useState<ProofBuildingStep[]>([]);
        const [activeStep, setActiveStep] = useState<number>(-1);
    
        const [foundTransitions, setFoundTransitions] = useState<string[]>([]);

        // load state from persisted state
        const [stateLoaded, setStateLoaded] = useState<boolean>(false);
        useEffect(() => {
            if (stateLoaded) return;
            setStateLoaded(true); // set this early – we don't want to re-try even if we fail (so a corrupt state just means we start at the default state)

            if (!state) return;

            restoreInteractiveCCSProofTreeState(state, initialTree, context, setSteps, setActiveStep);

            const newFoundTransitions: string[] = [];
            if (state.foundTransitions instanceof Array) for (const t of state.foundTransitions) if (typeof t === "string") newFoundTransitions.push(t);
            setFoundTransitions(newFoundTransitions);
        }, [stateLoaded, state, initialTree, context]);
    
        // persist steps into the InteractiveComponent-provided state
        useEffect(() => {
            setState(objectifyInteractiveProofTreeState(steps, activeStep));
        }, [steps, activeStep, setState]);
    
        const onFinalProofChanged: Parameters<typeof InteractiveCCSProofTree>[0]["onFinalProofChanged"] = (finishedDerivation) => {
            if (finishedDerivation) {
                switch(solutionMode.type) {
                    case "finish-the-tree": {
                        markAsSolved();
                        break;
                    }
                    case "find-enough-transitions": {
                        const transition = `{${finishedDerivation.process.toTeX()}} \\xrightarrow{${finishedDerivation.getStep().action.toTeX()}} {${finishedDerivation.getStep().target.toTeX()}}`;
                        if (!foundTransitions.includes(transition)) {
                            const newFoundTransitions = [...foundTransitions, transition];
                            setFoundTransitions(newFoundTransitions);
                            if (newFoundTransitions.length >= solutionMode.minTransitionsToFind) markAsSolved();
                        }    
                        break;
                    }
                }
            }
        };

        useEffect(() => {
            if (solutionMode.type === "reach-number-of-steps") {
                let eligibleSteps = steps;
                if (solutionMode.onlyCountRuleApplications) eligibleSteps = eligibleSteps.filter((s) => s.action instanceof RuleApplicationProofBuildingAction && solutionMode.onlyCountRuleApplications?.includes(s.action.ruleId));
                if (eligibleSteps.length >= solutionMode.stepCount) markAsSolved();
            }
        });

        const onApplyRuleModalOpened = () => {
            if (solutionMode.type === "open-the-rule-picker") markAsSolved();
        };

        const onAttemptToRevealFailedRuleApplication = () => {
            if (solutionMode.type === "attempt-to-reveal-failed-rule-application") markAsSolved();
        };

        return <Container fluid className="py-3 ccs-doctor">
            { customInstruction ? <div>{ customInstruction }</div> : solutionMode.type === "finish-the-tree" ? <div>
                Complete the following inference tree:
            </div> : solutionMode.type === "find-enough-transitions" ? <div>
                To solve this exercise, you must derive { solutionMode.minTransitionsToFind } transition{ solutionMode.minTransitionsToFind !== 1 ? "s" : "" }.
                { foundTransitions.length > 0 ? <>
                    {" "}Already found: <div>
                        <ul>
                            { foundTransitions.map((t, i) => <li key={i}><MathSpan content={t} /></li>) }
                        </ul>
                        Use the breadcrumbs below to return to previous steps.
                    </div>
                </> : null }
            </div> : null }
            <InteractiveCCSProofTree
                context={context}
                initialTree={initialTree}
                steps={steps} setSteps={setSteps}
                activeStep={activeStep} setActiveStep={setActiveStep}
                enabledRuleSet={enabledRuleSet}
                onFinalProofChanged={onFinalProofChanged}
                onApplyRuleModalOpened={onApplyRuleModalOpened}
                onAttemptToRevealFailedRuleApplication={onAttemptToRevealFailedRuleApplication} />
        </Container>;
    };
    
    return CCSProofTreeExercise;
};

export default makeCCSProofTreeExercise;
