import React, { useCallback, useContext, useEffect, useState } from "react";
import { useMemo } from "react";
import { InteractiveComponent } from "../../../hybrid-documents-reader-core/interactive/InteractiveElement";
import { CCSRuleSet, InteractiveCCSProofTree, ProofBuildingStep, ProofTreeAutoRevealStrategy } from "../../../pseuco-shared-components/ui/ccsDerivation/InteractiveProofTree";
import { LTSViewData } from "../../../pseuco-shared-components/ui/editors/lts/LTSInteractiveView";
import MathSpan from "../../../pseuco-shared-components/ui/mathSpan/MathSpan";

import "./ccsSemantics.scss";
import { Button, Modal } from "react-bootstrap";
import LTSSVGFrame from "../../../pseuco-shared-components/ui/editors/lts/LTSSVGFrame";
import LTSSVGFragmentView from "../../../pseuco-shared-components/ui/editors/lts/LTSSVGFragmentView";
import { CurrentThemeContext } from "../../../hybrid-documents-reader-core/ThemeApplicator";
import { getMaxLTSCoordinate } from "../DoubleSidedLTS";
import CenterOrShrinkGraphics from "../CenterOrShrinkGraphics";
import { CCSProcess, CCSSatisfiedRuleApplication, CCSSemanticContext, CCSUnfulfilledPremise, parseCCS, parseCCSAction, parseCCSProcess } from "@pseuco/ccs-interpreter";
import { ToastContext } from "../../../hybrid-documents-reader-core/Toaster";

/**
 * Used to safely assert non-nullness when using a state update in a modal.
 */
const notNull = <T,>(v: T | null): T => {
    if (v === null) throw new Error("A modal state variable was null while trying to process an update to it.");
    return v;
};

/**
 * Turns a transition into a key used to persist the reveal state.
 */
const codifyTransition = (source: string, action: string, target: string) => `〈${source}〉${action}〈${target}〉`;

const makeCCSSemantics = ({
    lts,
    ccs: ccsString,
    enabledRuleSet,
    proofTreeAutoRevealStrategy,
    prefillRevealModalFromPremise
}: {
    lts: LTSViewData,
    ccs: string,
    enabledRuleSet?: CCSRuleSet,
    proofTreeAutoRevealStrategy?: ProofTreeAutoRevealStrategy,
    prefillRevealModalFromPremise?: boolean;
}) => {
    const ccs = parseCCS(ccsString); // this one is important, as it also contains the process definitions
    const context = new CCSSemanticContext(ccs.environment);

    const CCSSemantics: InteractiveComponent = ({ state, setState, markAsSolved }) => {
        const initialCCSString = useMemo(() => {
            const initial = Object.entries(lts.states).find(([, s]) => s.initial);
            if (!initial) throw new Error("Failed to find the initial state.");
            return initial[0];
        }, []);
    
        const initialCCS = useMemo(() => parseCCS(initialCCSString), [initialCCSString]);

        const theme = useContext(CurrentThemeContext);
        const toaster = useContext(ToastContext);

        const [showHelpModal, setShowHelpModal] = useState<boolean>(false);

        const [investigateStateSectionState, setInvestigateStateSectionState] = useState<{
            show: boolean;
            stateProcess: CCSProcess;
            stateKey: string;
            initialTree: CCSUnfulfilledPremise;
            steps: ProofBuildingStep[];
            activeStep: number;
            didAddTransition: boolean;
        } | null>(null);

        const [revealedTransitions, setRevealedTransitions] = useState<string[]>([]);
        
        const [dataFromStateLoaded, setDataFromStateLoaded] = useState<boolean>(false);
        useEffect(() => {
            if (dataFromStateLoaded) return;
            setDataFromStateLoaded(true);
            if (!state) return;
            const transitions = [];
            try {
                const revealedTransitionsFromState: unknown = state.revealedTransitions;
                if (revealedTransitionsFromState instanceof Array) {
                    for (const t of revealedTransitionsFromState) {
                        if (typeof t === "string" && !revealedTransitions.includes(t)) transitions.push(t);
                    }
                }
            } catch (e) {
                console.log("Unexpected error when restoring state. Ignoring saved data.");
                console.log(e);
            }
            setRevealedTransitions(transitions);
        }, [dataFromStateLoaded, revealedTransitions, state]);

        useEffect(() => {
            setState({ revealedTransitions });
        }, [revealedTransitions, setState]);

        const [visibleStates, visibleTransitions]: [LTSViewData["states"], LTSViewData["transitions"]] = useMemo(() => {
            const states: LTSViewData["states"] = {};
            const transitions: LTSViewData["transitions"] = [];

            // initial state is always visible
            Object.entries(lts.states).forEach(([ccs, s]) => { if (s.initial) states[ccs] = s; });

            lts.transitions.forEach((t) => {
                const key = codifyTransition(t.source, t.label, t.target);
                if (revealedTransitions.includes(key)) {
                    transitions.push(t);
                    if (!(t.target in states)) states[t.target] = lts.states[t.target];
                }
            });

            if (investigateStateSectionState) states[investigateStateSectionState.stateKey] = { ...states[investigateStateSectionState.stateKey], highlighted: true };

            return [states, transitions];
        }, [investigateStateSectionState, revealedTransitions]);

        const imageSize = useMemo(() => getMaxLTSCoordinate(lts), []);

        const onStateClick = useCallback((stateKey: string) => {
            if (investigateStateSectionState && stateKey === investigateStateSectionState.stateKey) {
                // deselection behavior
                setInvestigateStateSectionState(null);
                return;
            }

            let stateProcess: CCSProcess;
            try {
                stateProcess = parseCCSProcess(stateKey);
            } catch (e) {
                throw new Error(`An error occurred while parsing the CCS process ${stateKey}`);
            }

            setInvestigateStateSectionState({
                show: true,
                stateProcess,
                stateKey,
                initialTree: new CCSUnfulfilledPremise(stateProcess),
                steps: [],
                activeStep: -1,
                didAddTransition: false
            });
        }, [investigateStateSectionState]);

        const done = () => {
            if (visibleTransitions.length >= lts.transitions.length) {
                toaster.toast("success", "All Transitions Found", "Congratulations, you have found all transitions! Your LTS is complete.", 5000);
                markAsSolved();
            } else {
                toaster.toast("error", "Missing Transitions", "Your are still missing at least one transition. Carefully check all states again.", 10000);
            }
        };

        const reset = () => {
            setRevealedTransitions([]);
            setInvestigateStateSectionState(null);
            setShowResetModal(false);
        };

        const [showResetModal, setShowResetModal] = useState<boolean>(false);

        const investigateStateSection = investigateStateSectionState ? (() => {
            const revealedTransitionsFromThisState = visibleTransitions.flatMap((t) => t.source === investigateStateSectionState.stateKey ? [`{} \\xrightarrow{${parseCCSAction(t.label).toTeX()}} {${parseCCSProcess(t.target).toTeX()}}`] : []);
            
            const stateUnderInvestigationIsExemptByExercise = lts.states[investigateStateSectionState.stateKey].expandable;

            if (stateUnderInvestigationIsExemptByExercise) return <div>
                <p>Good news: You do not need to explore this state as part of this exercise.</p>
                <p>Make sure all other states are completed, then click <em>Done</em>.</p>
            </div>;

            const onFinalProofChanged = (finishedDerivation: CCSSatisfiedRuleApplication | null) => {
                if (!finishedDerivation) return;
                const step = finishedDerivation.getStep();
                const key = codifyTransition(step.source.toString(), step.action.toString(), step.target.toString());
                if (!revealedTransitions.includes(key)) {
                    // new transition discovered
                    setRevealedTransitions([...revealedTransitions, key]);
                    toaster.toast("success", "Transition Added", "This transition has been added to the LTS.", 5000);
                    setInvestigateStateSectionState({ ...investigateStateSectionState, didAddTransition: true });
                }
            };

            return <div>
                <p>You are currently considering the state <MathSpan content={investigateStateSectionState.stateProcess.toTeX()} />.</p>
                { revealedTransitionsFromThisState.length ? <>
                    <p>Known transitions:</p>
                    <ul>
                        { revealedTransitionsFromThisState.map((t, i) => <li key={i}><MathSpan content={t} /></li>) }
                    </ul>
                </> : null }
                <p>Complete the inference below to add a transition:</p>
                <InteractiveCCSProofTree
                    key={investigateStateSectionState.stateKey}
                    context={context}
                    initialTree={investigateStateSectionState.initialTree}
                    steps={investigateStateSectionState.steps} setSteps={(steps) => setInvestigateStateSectionState((s) => ({ ...notNull(s), steps }))}
                    activeStep={investigateStateSectionState.activeStep} setActiveStep={(activeStep) => setInvestigateStateSectionState((s) => ({ ...notNull(s), activeStep }))}
                    onFinalProofChanged={onFinalProofChanged}
                    enabledRuleSet={enabledRuleSet}
                    proofTreeAutoRevealStrategy={proofTreeAutoRevealStrategy}
                    prefillRevealModalFromPremise={prefillRevealModalFromPremise}
                />
            </div>;
        })() : null;

        return <div className="ccs-semantics-exercise">
            <Modal show={showHelpModal} onHide={() => setShowHelpModal(false)}>
                <Modal.Header closeButton>
                    <Modal.Title>Help</Modal.Title>
                </Modal.Header>
                <Modal.Body>
                    <p>In this type of exercise, your job is to construct an LTS from a CCS term using the CCS semantics.</p>
                    <p>At the start, you will be given a CCS term, and a single state will be shown in the LTS, representing this CCS term.</p>
                    <p>To add a transition from a state (or just to see the CCS term), simply click or tap a state. This will open a dialog that allows you to construct an inference tree justifying a transition.</p>
                    <p>Any transition that you have inferred is automatically added to the LTS, adding the target state if it is not already in the LTS. Once you think you have found all transitions, click the <em>Done</em> button in the corner to check your result.</p>
                </Modal.Body>
                <Modal.Footer>
                    <Button variant="primary" onClick={() => setShowHelpModal(false)}>Close</Button>
                </Modal.Footer>
            </Modal>
            <Modal show={showResetModal} onHide={() => setShowResetModal(false)}>
                <Modal.Header closeButton>
                    <Modal.Title>Reset Exercise</Modal.Title>
                </Modal.Header>
                <Modal.Body>
                    <p>Do you really want to reset this exercise? This action cannot be undone.</p>
                </Modal.Body>
                <Modal.Footer>
                    <Button variant="secondary" onClick={() => setShowResetModal(false)}>Cancel</Button>
                    <Button variant="danger" onClick={reset}>Reset Exercise</Button>
                </Modal.Footer>
            </Modal>
            <div className="clearfix">
                <span>Construct <MathSpan content={`\\llbracket ${initialCCS.system.toTeX()} \\rrbracket`} />!</span>
                <span className="control-buttons">
                    <Button size="sm" variant="secondary" onClick={() => setShowHelpModal(true)}>Show Help</Button>
                    <Button size="sm" variant="secondary" disabled={visibleTransitions.length === 0} onClick={() => setShowResetModal(true)}>Reset</Button>
                    <Button size="sm" variant="primary" onClick={done}>Done</Button>
                </span>
            </div>
            <div>
                <CenterOrShrinkGraphics width={imageSize.x + 20}>
                    <LTSSVGFrame darkMode={theme === 'dark'} hideWeak={false} largeFont={false} moreSvgProps={{ viewBox: `-10 -10 ${imageSize.x + 20} ${imageSize.y + 20}` }}>
                        <LTSSVGFragmentView states={visibleStates} transitions={visibleTransitions} showExpandNotice={false} onStateClick={onStateClick} />
                    </LTSSVGFrame>
                </CenterOrShrinkGraphics>
            </div>
            { investigateStateSection }
        </div>;
    };
    
    return CCSSemantics;
};

export default makeCCSSemantics;
