import React, { useContext, ReactElement, Fragment, PropsWithChildren } from 'react';
import { InteractiveComponent } from '../../../hybrid-documents-reader-core/interactive/InteractiveElement';
import { DoubleSided, prepareDoubleLTSForRendering, Side, otherSide, addShifts, shiftToTransform, isSide, Pair, sides } from '../DoubleSidedLTS';
import { LTSViewData } from '../../../pseuco-shared-components/ui/editors/lts/LTSInteractiveView';
import CenterOrShrinkGraphics from '../CenterOrShrinkGraphics';
import LTSSVGFrame from '../../../pseuco-shared-components/ui/editors/lts/LTSSVGFrame';
import { CurrentThemeContext } from '../../../hybrid-documents-reader-core/ThemeApplicator';

import "./disprovingBisimilarity.scss";
import { Button } from 'react-bootstrap';
import MathSpan from '../../../pseuco-shared-components/ui/mathSpan/MathSpan';
import { LTSTransitionViewData, XY } from '../../../pseuco-shared-components/ui/editors/lts/LTSSVGFragmentView';
import classNames from 'classnames';
import { LTSTransitionViewDataAlike, generateWeakTransitions } from '../ltsUtils';
import { mapObject } from '../../../hybrid-documents-reader-core/util/functional';

const throwIfNull = <T,>(value: T | null): T => {
    if (value === null) throw new Error("An expected value was null.");
    return value;
};

const UlOrNothing: React.FC<PropsWithChildren<{ ul: boolean }>> = ({ ul, children }) => ul ? <ul>{children}</ul> : <Fragment>{children}</Fragment>;

const LiOrNothing: React.FC<PropsWithChildren<{ li: boolean }>> = ({ li, children }) => li ? <li>{children}</li> : <Fragment>{children}</Fragment>;

const DisprovingBisimilarityBuilder = (lts: DoubleSided<LTSViewData>, initialTuple: DoubleSided<string>, {
    noIntermediateSideSwaps = false,
    defenderHeuristicRelation = undefined,
    weak = false,
    observationCongruent = false,
    weakAttackerSteps = false
}: {
    noIntermediateSideSwaps?: boolean;
    defenderHeuristicRelation?: DoubleSided<string>[];
    weak?: boolean;
    observationCongruent?: boolean;
    weakAttackerSteps?: boolean;
}): InteractiveComponent => {
    if (!weak) {
        if (weakAttackerSteps) throw new Error("Cannot use weak attacker steps when weak mode is off.");
        if (observationCongruent) throw new Error("Inconsistent options for disprovingBisimilarity interactive element.");
    }

    const innerMargin = 100;
    const margins = {
        top: 50, // to accommodate relation
        bottom: 20,
        left: 110,
        right: 110
    };
    const { doubleLtsElement, shifts, imageSize } = prepareDoubleLTSForRendering(lts, margins, innerMargin);
    
    const weakTransitions: DoubleSided<LTSTransitionViewDataAlike[]> = { left: [], right: [] };
    if (weak) {
        sides.forEach((side) => weakTransitions[side] = generateWeakTransitions(lts[side]));
    }

    /**
     * The transitions the attacker will use – normal ones or weak ones.
     */
    const attackerTransitions: DoubleSided<LTSTransitionViewDataAlike[]> = weakAttackerSteps ? weakTransitions : mapObject(lts, (lts) => lts.transitions) as DoubleSided<LTSTransitionViewData[]>; // cast needed because TypeScript does not understand mapObject keeps the keys

    const DisprovingBisimilarity: InteractiveComponent = ({ state, setState, markAsSolved }) => {
        const theme = useContext(CurrentThemeContext);
        
        const [delayDefenderStep, setDelayDefenderStep] = React.useState<boolean>(false);
        
        /**
         * Describes the decision taken at any point during the game and the follow-ups.
         */
        type AttackStrategyElement = {
            attackerDecision?: {
                side: Side;
                transition: number; // as an index into the LTS view data, or into weakTransitions
                followUps: AttackStrategyElement[]; // for every possible defense (generated in-order), the next element of the strategy
            };
        }

        /**
         * A fuller representation of the game than [[AttackStrategyElement]], containing convenience information that is not to be persisted.
         */
        type GameStep = {
            initial: boolean;
            pair: Pair;
            pathToHere: number[];
            attackerDecision?: {
                side: Side;
                transitionIndex: number;
                transition: LTSTransitionViewDataAlike;
                defenderOptions: {
                    side: Side;
                    transitionIndex: number;
                    transition: LTSTransitionViewDataAlike;
                    targetStep: GameStep;
                }[];
            };
        }

        /**
         * Determines whether a transition really exists on a specific side (vs. the transition being only a weak transition or being made up).
         * @param side The side to check.
         * @param transition The transision to check for.
         */
        const hasRealTransition = (side: Side, transition: LTSTransitionViewDataAlike): boolean => lts[side].transitions.some((t) => t.source === transition.source && t.target === transition.target && t.weak === transition.weak && t.label === transition.label);

        /**
         * Given an attack, returns defending transitions (and their indexes) that are valid and to be considered according to the exercise configuration. This is deterministic.
         * @param gameStep The current state of the game.
         * @param defendingSide Which side to defend on.
         * @param action Which action to defend against.
         */
        const getDefenderOptions = (gameStep: GameStep, defendingSide: Side, attackingTransition: LTSTransitionViewDataAlike): { transition: LTSTransitionViewDataAlike; index: number }[] => {
            const possibleDefenses = (weak ? weakTransitions[defendingSide] : lts[defendingSide].transitions)
                .map((transition, index) => ({ transition, index }))
                .filter(({ transition }) => transition.source === gameStep.pair[defendingSide] && transition.label === attackingTransition.label)
                .filter(({ transition }) => !observationCongruent || !gameStep.initial || !attackingTransition.weak || !hasRealTransition(otherSide(defendingSide), attackingTransition) || hasRealTransition(defendingSide, transition));

            if (defenderHeuristicRelation) {
                // we're configured not to consider all cases, only one

                const defensesMatchingStrategy = possibleDefenses.filter(({ transition }) => defenderHeuristicRelation.some((relationEntry) => relationEntry[defendingSide] === transition.target && relationEntry[otherSide(defendingSide)] === attackingTransition.target));

                if (defensesMatchingStrategy.length > 0) {
                    return [defensesMatchingStrategy[0]]; // take this first one that matches
                } else if (possibleDefenses.length > 0) {
                    return [possibleDefenses[0]]; // nothing matches, so take the first one
                } else {
                    return []; // nothing possible at all
                }
            } else {
                // do it properly
                return possibleDefenses;
            }
        };

        let attackerSide: Side | null = null;
        const attackStrategy: AttackStrategyElement = {};
        let currentPath: (number | null)[] = []; // null indicates the attacker won, and we are in the victory state

        const game: GameStep = {
            initial: true,
            pair: initialTuple,
            pathToHere: []
        };

        // attempt to restore state
        if (state) {
            if (isSide(state.attackerSide)) attackerSide = state.attackerSide;

            // restore attackStrategy and synthesize game
            const restoreStrategy = (sourceStrategyElement: AttackStrategyElement, targetStrategyElement: AttackStrategyElement, gameStep: GameStep): void => {
                if (typeof sourceStrategyElement === "object" && sourceStrategyElement !== null) {
                    const currentPair = gameStep.pair;
                    if (sourceStrategyElement.attackerDecision) {
                        const attackingSide = sourceStrategyElement.attackerDecision.side;
                        const defendingSide = otherSide(attackingSide);
                        if (!isSide(attackerSide)) return;
                        const attackingTransitionIndex = sourceStrategyElement.attackerDecision.transition;
                        const attackingTransition = attackerTransitions[attackingSide][attackingTransitionIndex];
                        if (!(attackingTransition && attackingTransition.source === currentPair[attackingSide])) return; // invalid attack

                        const followUps = sourceStrategyElement.attackerDecision.followUps;

                        // attack valid so far, check defenses
                        const defensiveTransitions = getDefenderOptions(gameStep, defendingSide, attackingTransition);
                        if (defensiveTransitions.length !== followUps.length) return;

                        const targetFollowUps = defensiveTransitions.map(() => ({}));
                        targetStrategyElement.attackerDecision = {
                            side: attackingSide,
                            transition: attackingTransitionIndex,
                            followUps: targetFollowUps
                        };

                        const newDefenderOptions = defensiveTransitions.map((defensiveTransition, index) => ({
                            side: defendingSide,
                            transitionIndex: defensiveTransition.index,
                            transition: defensiveTransition.transition,
                            targetStep: {
                                initial: false,
                                pair: { [attackingSide]: attackingTransition.target, [defendingSide]: defensiveTransition.transition.target } as Pair,
                                pathToHere: [...gameStep.pathToHere, index]
                            }
                        }));
                        gameStep.attackerDecision = {
                            side: attackingSide,
                            transitionIndex: attackingTransitionIndex,
                            transition: attackingTransition,
                            defenderOptions: newDefenderOptions
                        };

                        // now, recurse into the newly created attacker strategy element and game step
                        defensiveTransitions.forEach((defensiveTransition, i) => restoreStrategy(followUps[i], targetFollowUps[i], newDefenderOptions[i].targetStep));
                    }
                }
            };
            restoreStrategy(state.attackStrategy, attackStrategy, game);

            // restore path
            if (Array.isArray(state.currentPath)) {
                let currentStep = game;
                let i = 0;
                while (true) {
                    if (typeof state.currentPath[i] === "number" && currentStep.attackerDecision && currentStep.attackerDecision.defenderOptions[state.currentPath[i]]) {
                        // regular path element
                        currentPath.push(state.currentPath[i]);
                        currentStep = currentStep.attackerDecision.defenderOptions[state.currentPath[i]].targetStep;
                        i++;
                    } else if (state.currentPath[i] === null && currentStep.attackerDecision && currentStep.attackerDecision.defenderOptions.length === 0) {
                        // we are in a victory state
                        currentPath.push(null);
                        break;
                    } else {
                        break;
                    }
                }
            }
        }

        const getAttackStrategyAtPath = (path: (number | null)[]): AttackStrategyElement => {
            let result: AttackStrategyElement = attackStrategy;
            path.forEach((index) => {
                if (index === null) throw new Error("Trying to look up an attack strategy element for a success path.");
                const next = result.attackerDecision?.followUps[index];
                if (!next) throw new Error(`Failed to find an attack strategy element at path ${path}.`);
                result = next;
            });
            return result;
        };
        
        const getGameStepAtPath = (path: (number | null)[]): GameStep => {
            let result: GameStep = game;
            path.forEach((index) => {
                if (index === null) throw new Error("Trying to look up a game step for a success path.");
                const next = result.attackerDecision?.defenderOptions[index]?.targetStep;
                if (!next) throw new Error(`Failed to find a game step at path ${path}.`);
                result = next;
            });
            return result;
        };

        const currentlyInVictoryPosition = currentPath[currentPath.length - 1] === null;

        let currentGameStep: GameStep | null = null;
        let currentPositions: Pair | null = null;
        let attackerWinningPosition: string | null = null;
        if (currentlyInVictoryPosition) {
            const lastGameState = getGameStepAtPath(currentPath.slice(0, currentPath.length - 1));
            if (!lastGameState.attackerDecision) throw new Error("Internal inconsistency: In victory position, but previous step is undecided.");
            attackerWinningPosition = lastGameState.attackerDecision.transition.target;
        } else {
            currentGameStep = getGameStepAtPath(currentPath);
            currentPositions = currentGameStep.pair;
        }

        const saveState = (): void => {
            setState({
                attackerSide,
                attackStrategy,
                currentPath
            });    
        };    

        const resetState = (): void => setState(null);

        const setAttackingSide = (attackingSide: Side | null): void => {
            setDelayDefenderStep(false);
            attackerSide = attackingSide;
            saveState();
        };

        const jumpToPath = (path: number[]): void => {
            setDelayDefenderStep(false);
            currentPath = path;
            saveState();
        };

        const onTransitionClick = (side: Side, transitionId: number): void => {
            if (side !== attackerSide) throw new Error("Attacking on the wrong side!");
            if (!currentGameStep) throw new Error("Trying to attack in a victory state!");
            const currentAttackStrategyElement = getAttackStrategyAtPath(currentPath);
            const transition = attackerTransitions[side][transitionId];

            const followUps = getDefenderOptions(currentGameStep, otherSide(attackerSide), transition).map(() => ({}));
            currentAttackStrategyElement.attackerDecision = {
                side: attackerSide,
                transition: transitionId,
                followUps
            };
            currentPath.push(followUps.length > 0 ? 0 : null);
            setDelayDefenderStep(true);
            saveState();
        };

        const defenderSide: Side | null = attackerSide === null ? null : otherSide(attackerSide);

        const possibleAttacks = attackerSide === null || currentGameStep === null ? [] : attackerTransitions[attackerSide].map((transition, index) => ({ transition, index })).filter(({ transition }) => transition.source === throwIfNull(currentGameStep).pair[throwIfNull(attackerSide)]);

        const gameStepCompleted = (gameStep: GameStep): boolean => !!gameStep.attackerDecision && gameStep.attackerDecision.defenderOptions.every((defenderOption) => gameStepCompleted(defenderOption.targetStep));
        const attackCompleted = gameStepCompleted(game);

        if (attackCompleted) markAsSolved();

        const renderTransitionClickTargets = (): ReactElement[] => {
            if (weakAttackerSteps) {
                return []; // weak transitions, so nothing to show in LTS
            } else {
                return possibleAttacks.map(({ transition, index }) => {
                    const t = transition as LTSTransitionViewData;
                    if (!(typeof t.x === "number" && typeof t.y === "number")) throw new Error("Despite not using weak attacks, an attack failed to have coordinates!");
                    return <circle className="click-target" cx={t.x} cy={t.y} key={index} r="20" onClick={(): void => onTransitionClick(throwIfNull(attackerSide), index)} />;
                });
            }
        };

        const renderPositionHighlight = (side: Side, isAttacker: boolean, shift: XY): ReactElement => <g className={`position-highlight ${isAttacker ? "attacker" : "defender"}`} style={{ transform: shiftToTransform(shift) }}>
            <circle r="30" fill={ isAttacker ? "red" : "blue" } />
            <text x={ side === "left" ? -40 : 40 } textAnchor={ side === "left" ? "end" : "start" } stroke={ isAttacker ? "red" : "blue" } dy="0.3em">{ side === "right" ? "← " : null }{ isAttacker ? "Attacker" : "Defender" }{ side === "left" ? " →" : null }</text>
        </g>;

        const getArrow = (weak: boolean): string => weak ? "\\xRightarrow" : "\\xrightarrow";

        const renderTransition = (transition: LTSTransitionViewDataAlike, weak: boolean): ReactElement => <MathSpan content={`\\mathit{${transition.source}} ${getArrow(weak)}{\\mathit{${transition.label}}} \\mathit{${transition.target}}`} />;

        const renderGameStep = (gameStep: GameStep): ReactElement => <Fragment>
            We are in states <MathSpan content={`\\mathit{${gameStep.pair.left}}`} /> and <MathSpan content={`\\mathit{${gameStep.pair.right}}`} />.
            { JSON.stringify(currentPath) === JSON.stringify(gameStep.pathToHere) ? <span className="you-are-here"> (You are here. You can select an attack above.)</span> : <span className="jump-to-here"> <Button size="sm" variant={ gameStep.attackerDecision ? "secondary" : "primary" } onClick={(): void => jumpToPath(gameStep.pathToHere)}>Select this case</Button></span> }
            { gameStep.attackerDecision ? <Fragment>
                {" "}The attacker picks the { gameStep.attackerDecision.side } LTS and moves { renderTransition(gameStep.attackerDecision.transition, weakAttackerSteps)}. { gameStep.attackerDecision.defenderOptions.length > 0 ? <Fragment>
                    { defenderHeuristicRelation ? <br /> : <Fragment>The defender can pick the following moves:</Fragment> }
                    <UlOrNothing ul={!defenderHeuristicRelation}>
                        { gameStep.attackerDecision.defenderOptions.map((defenderOption, i) => <LiOrNothing key={i} li={!defenderHeuristicRelation}>
                                The defender moves { renderTransition(defenderOption.transition, weak) }.{" "}
                            { renderGameStep(defenderOption.targetStep) }
                        </LiOrNothing>) }
                    </UlOrNothing>
                </Fragment> : <Fragment>
                    The defender cannot defend! ✓
                    { JSON.stringify(currentPath) === JSON.stringify([...gameStep.pathToHere, null]) ? <span className="you-are-here"> (You are here.)</span> : null }
                </Fragment>
                }
            </Fragment> : null }
        </Fragment>;

        return <div className={classNames({ "disproving-bisimilarity": true, "delay-defender-step": delayDefenderStep })}>
            <CenterOrShrinkGraphics width={imageSize.x}>
                <LTSSVGFrame darkMode={theme === 'dark'} hideWeak={false} largeFont={false} moreSvgProps={{ viewBox: `0 0 ${imageSize.x} ${imageSize.y}` }}>
                    { attackerSide === null ? null : renderPositionHighlight(attackerSide, true, addShifts(shifts[attackerSide], lts[attackerSide].states[throwIfNull(currentPositions === null ? attackerWinningPosition : currentPositions[attackerSide])])) }
                    { defenderSide === null || currentPositions === null ? null : renderPositionHighlight(defenderSide, false, addShifts(shifts[defenderSide], lts[defenderSide].states[currentPositions[defenderSide]])) }
                    { doubleLtsElement }
                    { attackerSide === null ? null : <g style={{ transform: shiftToTransform(shifts[attackerSide]) }}>
                        { renderTransitionClickTargets() }
                    </g> }
                </LTSSVGFrame>
            </CenterOrShrinkGraphics>
            <div>
                { attackerSide === null ?
                    <p>
                        You have not picked a side yet.
                        <Button className="action-button" onClick={(): void => setAttackingSide("left")}>Attack in left LTS</Button>
                        <Button className="action-button" onClick={(): void => setAttackingSide("right")}>Attack in right LTS</Button>
                    </p>
                    :
                    <p>
                        { currentPositions === null ? <Fragment>
                            The attacker has won this scenario! Check for other paths below.
                        </Fragment> : possibleAttacks.length > 0 ? <Fragment>
                            Attack by { weakAttackerSteps ? <span>selecting a weak transition below</span> : <span>clicking or tapping (the label of) a transition starting in the attacker's current state <MathSpan content={`\\mathit{${currentPositions[attackerSide]}}`} /></span> } { noIntermediateSideSwaps && currentPath.length !== 0 ? "." : <Fragment>, or <Button onClick={(): void => setAttackingSide(attackerSide ? otherSide(attackerSide) : null)}>Swap Sides</Button></Fragment> }
                            { weakAttackerSteps ? <span><br />
                                { possibleAttacks.map(({ transition: { source, label, target }, index }) => <Button className="weak-attack" key={index} onClick={(): void => onTransitionClick(throwIfNull(attackerSide), index)}>Attack with <MathSpan content={`\\mathit{${source}} \\xRightarrow{\\mathit{${label}}} \\mathit{${target}}`} /></Button>)}
                            </span> : null }
                        </Fragment> : <Fragment>
                            Oh no! There is no transition from the attacker's current state <MathSpan content={`\\mathit{${currentPositions[attackerSide]}}`} />.{ noIntermediateSideSwaps ? <Fragment> Go back using the buttons in the proof below.</Fragment> : <Fragment> You can try to <Button onClick={(): void => setAttackingSide(attackerSide ? otherSide(attackerSide) : null)}>Swap Sides</Button> or go back using the buttons in the proof below.</Fragment> }
                        </Fragment> }
                    </p>
                }
            </div>
            <div>
                { defenderHeuristicRelation ? <p>Let's play the game:</p> : <p>We prove the states are not { weak ? "weakly " : "" }bisimilar by playing the game:</p> }
                { renderGameStep(game) }
            </div>
            { attackCompleted ? <div>
                That's it! { defenderHeuristicRelation ? null : <Fragment>We've shown that the attacker can always win, therefore, we know the two LTS are not { weak ? "weakly " : "" }bisimilar.</Fragment> }
            </div> : null }
            { attackerSide === null ? null : <div>
                <p>
                    Want to start over?
                    <Button className="action-button" variant="danger" onClick={(): void => resetState()}>Reset Exercise</Button>
                </p>
            </div> }
        </div>;
    };

    return DisprovingBisimilarity;
};

export default DisprovingBisimilarityBuilder;
