import { InteractiveComponent } from '../../../hybrid-documents-reader-core/interactive/InteractiveElement';
import "./isomorphismRoulette.scss";
import { LTSViewData } from '../../../pseuco-shared-components/ui/editors/lts/LTSInteractiveView';
import React, { useContext, ReactElement, Fragment, useState, useEffect } from 'react';
import { DoubleSided, prepareDoubleLTSForRendering, Side, addShifts, shiftToTransform, shiftToCoords } from '../DoubleSidedLTS';
import LTSSVGFrame from '../../../pseuco-shared-components/ui/editors/lts/LTSSVGFrame';
import CenterOrShrinkGraphics from '../CenterOrShrinkGraphics';
import { CurrentThemeContext } from '../../../hybrid-documents-reader-core/ThemeApplicator';
import { mapObjectToArray } from '../../../pseuco-shared-components/util/functional';
import { Button } from 'react-bootstrap';
import { delay } from '../../../hybrid-documents-reader-core/util/async';
import classNames from 'classnames';
import { UserStudyContext } from '../../study/UserStudy';

const bezierShift = { x: 0, y: -80 };

/**
 * Creates an instance of the isomorphism roulette.
 * @param ltsBuckets An array of groups that are similar enough that the user could be asked about isomorphism. Each group consists of an array of buckets, each of which may not be isomorphic to the others. Each bucket is an array of isomorphic LTS.
 */
const IsomorphismRouletteBuilder = (ltsBuckets: LTSViewData[][][]): InteractiveComponent => {
    const IsomorphismRoulette: InteractiveComponent = ({ state, setState, markAsSolved, id: interactiveId }) => {
        const userStudy = useContext(UserStudyContext);

        type Exercise = {
            lts: DoubleSided<LTSViewData>;
            isomorphic: boolean;
        };

        let correctAnswers = 0;
        const correctAnswersNeeded = 10;
        if (state && typeof state === "object" && typeof state.correctAnswers === "number") correctAnswers = Math.floor(state.correctAnswers);

        const setCorrectAnswers = (count: number): void => {
            correctAnswers = count;
            setState({
                correctAnswers
            });
        };

        useEffect(() => {
            userStudy.reportEvent({ type: "INTERACTIVE_PROGRESS_UPDATE", interactiveType: "interactive", interactiveId, stage: 0, progress: Math.min(correctAnswersNeeded, correctAnswers) });
        }, [correctAnswers, interactiveId, userStudy]);

        const generateExercise = (): Exercise => {
            // begin by selecting the answer, so it's truly fifty-fifty
            const isomorphic = Math.random() >= 0.5;

            const bucketGroupIndex = Math.floor(Math.random() * ltsBuckets.length);
            const bucketGroup = ltsBuckets[bucketGroupIndex];

            if (bucketGroup.length < 2) throw new Error("Too few buckets!");
            
            const firstBucketIndex = Math.floor(Math.random() * bucketGroup.length);
            const firstBucket = bucketGroup[firstBucketIndex];

            const firstLtsIndex = Math.floor(Math.random() * firstBucket.length);
            const firstLts = firstBucket[firstLtsIndex];

            if (firstBucket.length < 2) throw new Error("Too few examples in bucket!");

            // now pick the second LTS appropriately
            if (isomorphic) {
                let secondLtsIndex = firstLtsIndex;
                while (secondLtsIndex === firstLtsIndex) secondLtsIndex = Math.floor(Math.random() * firstBucket.length);
                const secondLts = firstBucket[secondLtsIndex];

                return {
                    lts: { left: firstLts, right: secondLts },
                    isomorphic
                };
            } else {
                let secondBucketIndex = firstBucketIndex;
                while (secondBucketIndex === firstBucketIndex) secondBucketIndex = Math.floor(Math.random() * bucketGroup.length);
                const secondBucket = bucketGroup[secondBucketIndex];
                const secondLtsIndex = Math.floor(Math.random() * secondBucket.length);
                const secondLts = secondBucket[secondLtsIndex];

                return {
                    lts: { left: firstLts, right: secondLts },
                    isomorphic
                };
            }
        };

        const [exercise, setExercise] = useState<Exercise>(generateExercise());
        const [showResult, setShowResult] = useState<boolean>(false);
        const [lastResult, setLastResult] = useState<boolean>(false);

        const [selection, setSelection] = useState<null | { side: Side; stateKey: string }>(null);
        const [relation, setRelation] = useState<DoubleSided<string>[]>([]);

        const theme = useContext(CurrentThemeContext);
        
        const innerMargin = 150;
        const margins = {
            top: 50, // to accommodate relation
            bottom: 20,
            left: 20,
            right: 150
        };
        const { doubleLtsElement, shifts, imageSize } = prepareDoubleLTSForRendering(exercise.lts, margins, innerMargin);

        const labelShift = {
            x: 30,
            y: 4
        };

        const stateClick = (side: Side, stateKey: string): void => {
            if (selection && selection.side === side && selection.stateKey === stateKey) {
                // clicked again, de-select
                setSelection(null);
            } else if (selection && selection.side !== side) {
                // make connection
                setRelation([...relation.filter((entry) => entry[side] !== stateKey), { [side]: stateKey, [selection.side]: selection.stateKey } as DoubleSided<string>]);
                setSelection(null);
            } else {
                setSelection({ side, stateKey });
                setRelation(relation.filter((entry) => entry[side] !== stateKey));
            }
        };

        const getStateLabels = (side: Side): ReactElement => <Fragment>
            { mapObjectToArray(exercise.lts[side].states, (state, stateKey) => stateKey.length > 2 ? <text key={stateKey} className="state-long-label" style={{ transform: shiftToTransform(addShifts(shifts[side], labelShift, state))}}>{ stateKey }</text> : null) };
        </Fragment>;

        const getStateClickTargets = (side: Side): ReactElement => <Fragment>
            { mapObjectToArray(exercise.lts[side].states, (state, stateKey) => <circle className="click-target" key={stateKey} r="30" onClick={(): void => stateClick(side, stateKey)} style={{ transform: shiftToTransform(addShifts(shifts[side], state))}} />) };
        </Fragment>;

        const getHighlight = (): ReactElement | null => selection ? <circle className="highlight" r="30" style={{ transform: shiftToTransform(addShifts(shifts[selection.side], exercise.lts[selection.side].states[selection.stateKey]))}} /> : null;

        const submitAnswer = (answer: boolean): void => {
            if (!showResult) {
                (async (): Promise<void> => {
                    const correct = answer === exercise.isomorphic;

                    if (!correct) userStudy.reportEvent({ type: "INTERACTIVE_REPORT_ERROR", interactiveType: "interactive", interactiveId });

                    setShowResult(true);
                    setLastResult(correct);
                    await delay(1000);

                    setShowResult(false);
                    setSelection(null);
                    setRelation([]);
                    setExercise(generateExercise());
                    if (correct) {
                        setCorrectAnswers(correctAnswers + 1);   
                        if (correctAnswers >= correctAnswersNeeded) markAsSolved();
                    } else {
                        setCorrectAnswers(0);
                    }
                })();
            }
        };

        const correctAnswersMissing = correctAnswersNeeded - correctAnswers;

        return <div className="isomorphism-roulette">
            <div className={classNames({
                result: true,
                active: showResult
            })}>
                {
                    lastResult ? 
                        <div className="icon success">
                        ✓
                        </div> : <div className="icon failure">
                        ✗
                        </div>
                }
            </div>
            <div className="content">        
                <CenterOrShrinkGraphics width={imageSize.x}>
                    <LTSSVGFrame darkMode={theme === 'dark'} hideWeak={true} largeFont={false} moreSvgProps={{ viewBox: `0 0 ${imageSize.x} ${imageSize.y}` }}>
                        { getHighlight() }
                        <g>
                            {relation.map(({ left, right }, i) => <path
                                key={i}
                                className="relation-tuple"
                                d={`M ${shiftToCoords(addShifts(shifts.left, exercise.lts.left.states[left]))} C ${shiftToCoords(addShifts(shifts.left, exercise.lts.left.states[left], bezierShift))}, ${shiftToCoords(addShifts(shifts.right, exercise.lts.right.states[right], bezierShift))}, ${shiftToCoords(addShifts(shifts.right, exercise.lts.right.states[right]))}`}
                            ></path>)}
                        </g>
                        { doubleLtsElement }
                        <g>
                            { getStateLabels("left") }
                            { getStateLabels("right") }
                        </g>
                        <g>
                            { getStateClickTargets("left") }
                            { getStateClickTargets("right") }
                        </g>
                    </LTSSVGFrame>
                </CenterOrShrinkGraphics>
                <p>
                    Are these LTS isomorphic?
                    <Button className="answer-button" onClick={(): void => submitAnswer(true)}>Yes</Button>
                    <Button className="answer-button" onClick={(): void => submitAnswer(false)}>No</Button>
                </p>
                <p>
                    You can click or tap states above to connect them to visualize the bijection. However, this is for your notes only – only your answer is counted.
                </p>
                <p>
                    { correctAnswers >= correctAnswersNeeded ? <Fragment>
                        Well done! You've completed this exercise.
                    </Fragment> : <Fragment>
                        Answer correctly { correctAnswers === 0 ? <Fragment>ten times in a row</Fragment> : <Fragment>{ correctAnswersMissing } more time{ correctAnswersMissing > 1 ? "s" : "" }</Fragment>}  to complete this exercise!
                    </Fragment> }
                </p>
            </div>
        </div>;
    };

    return IsomorphismRoulette;
};

export default IsomorphismRouletteBuilder;
