import React, { useContext, ReactElement, Fragment, useState, useEffect } from "react";
import { LTSViewData } from "../../../pseuco-shared-components/ui/editors/lts/LTSInteractiveView";
import { InteractiveComponent } from "../../../hybrid-documents-reader-core/interactive/InteractiveElement";

import "./minimization.scss";
import CenterOrShrinkGraphics from "../CenterOrShrinkGraphics";
import LTSSVGFrame from "../../../pseuco-shared-components/ui/editors/lts/LTSSVGFrame";
import LTSSVGFragmentView, { LTSTransitionViewData } from "../../../pseuco-shared-components/ui/editors/lts/LTSSVGFragmentView";
import { CurrentThemeContext } from "../../../hybrid-documents-reader-core/ThemeApplicator";
import { mapObject, mapObjectToArray, flatten } from "../../../pseuco-shared-components/util/functional";
import { getMaxLTSCoordinate, shiftToTransform } from "../DoubleSidedLTS";
import { Button, Modal, ButtonGroup } from "react-bootstrap";
import MathSpan from "../../../pseuco-shared-components/ui/mathSpan/MathSpan";
import { generateWeakTransitions, LTSTransitionViewDataAlike } from "../ltsUtils";
import { union } from "../../../hybrid-documents-reader-core/util/functional";
import { UserStudyContext } from "../../study/UserStudy";

const colors = [
    "#F20D0D",
    "#930DF2",
    "#43F20D",
    "#0DCAF2",
    "#F20DAF",
    "#F25E0D",
    "#E5F20D",
    "#0DF2C9",
    "#F2AF0D",
    "#808080"
];

const MinimizationBuilder = async (lts: LTSViewData, {
    preferredRepresentatives = [],
    transitionPositionOverrides = [],
    weak = false,
    observationCongruent = false
}: {
    preferredRepresentatives?: string[];
    transitionPositionOverrides?: {
        source: string;
        target: string;
        label: string;
        x: number;
        y: number;
    }[];
    weak?: boolean;
    observationCongruent?: boolean;
}): Promise<InteractiveComponent> => {
    if (observationCongruent && !weak) throw new Error("Inconsistent options for minimization interactive element.");

    const imageMargin = 15;
    const measurements = getMaxLTSCoordinate(lts);
    const imageSize = mapObject(measurements, (s) => s + 2 * imageMargin);

    const actions = Array.from(new Set(lts.transitions.map((t) => t.label))).sort();

    const weakTransitions: LTSTransitionViewDataAlike[] = weak ? generateWeakTransitions(lts) : [];

    const Minimization: InteractiveComponent = ({ state, setState, markAsSolved, id }) => {
        const theme = useContext(CurrentThemeContext);
        const userStudy = useContext(UserStudyContext);

        const statePartitions: { [stateKey: string]: number } = mapObject(lts.states, () => 0);
        let currentSplitter: null | {
            action: string;
            partition: number;
        } = null;

        const [highlighted, setHighlighted] = useState<string[]>([]);
        
        const [actionForDrawing, setActionForDrawing] = useState<string | null>(null);

        const [selectedTransitionSource, setSelectedTransitionSource] = useState<string | null>();

        let partitioningInitialized = false;
        let partitioningFinalized = false;
        let reconstructionFinalized = false;
        let transitionCleanupFinalized = false;
        let tauLoopInsertionFinalized = false;

        const mendedTransitionSelections: number[] = []; // indices of transitions that are copied to the minimized LTS

        let mendedTransitionNecessityCheckPosition = 0; // index into mendedTransitionSelections (not a transition id!) of the next transition to be checked, or mendedTransitionSelections.length when checking is done

        let addTauLoop = false;

        if (state && typeof state === "object") {
            if (typeof state.partitioningInitialized === "boolean") partitioningInitialized = state.partitioningInitialized;
            if (partitioningInitialized) {
                if (typeof state.partitioningFinalized === "boolean") partitioningFinalized = state.partitioningFinalized;
                if (state.statePartitions && typeof state.statePartitions === "object") {
                    Object.keys(statePartitions).forEach((stateKey) => {
                        if (typeof state.statePartitions[stateKey] === "number") statePartitions[stateKey] = state.statePartitions[stateKey];
                    });
                }
                if (partitioningFinalized) {
                    if (state.mendedTransitionSelections && Array.isArray(state.mendedTransitionSelections)) state.mendedTransitionSelections.forEach((selection: unknown) => {
                        if (typeof selection === "number") mendedTransitionSelections.push(selection);
                    });
                    if (typeof state.reconstructionFinalized === "boolean") reconstructionFinalized = state.reconstructionFinalized;
                    if (reconstructionFinalized) {
                        if (typeof state.mendedTransitionNecessityCheckPosition === "number" && state.mendedTransitionNecessityCheckPosition >= 0 && mendedTransitionNecessityCheckPosition <= mendedTransitionSelections.length) mendedTransitionNecessityCheckPosition = state.mendedTransitionNecessityCheckPosition;
                        if (typeof state.transitionCleanupFinalized === "boolean") transitionCleanupFinalized = state.transitionCleanupFinalized;
                        if (transitionCleanupFinalized) {
                            if (typeof state.tauLoopInsertionFinalized === "boolean") tauLoopInsertionFinalized = state.tauLoopInsertionFinalized;
                            if (typeof state.addTauLoop === "boolean") addTauLoop = state.addTauLoop;
                        }
                    }
                } else {
                    if (state.currentSplitter && typeof state.currentSplitter === "object" && typeof state.currentSplitter.action === "string" && typeof state.currentSplitter.partition === "number") currentSplitter = state.currentSplitter;
                }
            }
        }

        const saveState = (): void => {
            setState({
                statePartitions,
                partitioningInitialized,
                currentSplitter,
                partitioningFinalized,
                mendedTransitionSelections,
                reconstructionFinalized,
                mendedTransitionNecessityCheckPosition,
                transitionCleanupFinalized,
                tauLoopInsertionFinalized,
                addTauLoop
            });
        };

        const resetState = (): void => {
            setState(null);
            setHighlighted([]);
            setActionForDrawing(null);
        };

        // computed properties

        const partitions = Array.from(new Set(Object.values(statePartitions)));
        const partitionStates: { [partition: number]: string[] } = Object.fromEntries(partitions.map((partition) => [partition, Object.keys(lts.states).filter((stateKey) => statePartitions[stateKey] === partition).sort()])); // guaranteed to be sorted

        const splitResults: {
            [splitter: number]: {
                [action: string]: {
                    generatesSplit: boolean;
                    hitStates: string[];
                };
            };
        } = {};
        partitions.forEach((partition) => {
            splitResults[partition] = {};
            actions.forEach((action) => {
                const hitStates = [...new Set((weak ? weakTransitions : lts.transitions).filter(({ target, label }) => statePartitions[target] === partition && label === action).map(({ source }) => source))];

                let generatesSplit = false;
                partitions.forEach((targetPartition) => {
                    const hitAll = partitionStates[targetPartition].every((stateKey) => hitStates.includes(stateKey));
                    const hitNone = partitionStates[targetPartition].every((stateKey) => !hitStates.includes(stateKey));
                    if (!(hitAll || hitNone)) generatesSplit = true;
                });

                splitResults[partition][action] = {
                    generatesSplit,
                    hitStates
                };
            });
        });

        const representativeStateKeys = Object.fromEntries(partitions.map((partition) => {
            const candidateStates = partitionStates[partition];
            const preferredCandidateStates = candidateStates.filter((stateKey) => preferredRepresentatives.includes(stateKey));
            return [partition, preferredCandidateStates.length > 0 ? preferredCandidateStates[0] : candidateStates[0]];
        })); // maps partition number to the states that will survive minimization as the representatives of the partitions
        const representativeStates = Object.fromEntries(mapObjectToArray(representativeStateKeys, ((stateKey) => {
            const initial = partitionStates[statePartitions[stateKey]].some((sK) => lts.states[sK].initial);
            return [stateKey, { ...lts.states[stateKey], initial }];
        }))); // states object for the minimized LTS

        const getRepresentativeState = (stateKey: string): string => representativeStateKeys[statePartitions[stateKey]]; // maps a state to the representative

        const mendedTransitions: LTSTransitionViewData[] = lts.transitions.map((transition) => {
            const newSource = getRepresentativeState(transition.source);
            const newTarget = getRepresentativeState(transition.target);

            const override = transitionPositionOverrides.find(({ source, target, label }) => source === newSource && target === newTarget && label === transition.label);
            const newPosition = { x: transition.x, y: transition.y };
            if (override) {
                newPosition.x = override.x;
                newPosition.y = override.y;
            }

            return {
                ...transition,
                ...newPosition,
                source: newSource,
                target: newTarget,
            };
        });

        const selectedMendedTransitions: LTSTransitionViewData[] = [];
        mendedTransitionSelections.forEach((selection) => {
            if (mendedTransitions[selection]) selectedMendedTransitions.push(mendedTransitions[selection]);
        });

        const currentlyInTransitionNecessityCheck = weak && reconstructionFinalized && mendedTransitionNecessityCheckPosition < mendedTransitionSelections.length;

        const currentlyVisibleMendedTransitions = [...selectedMendedTransitions];
        if (currentlyInTransitionNecessityCheck) currentlyVisibleMendedTransitions.splice(mendedTransitionNecessityCheckPosition, 1); // hide the transition under investigation

        if (addTauLoop) {
            const originalInitialStateKey = Object.keys(lts.states).find((key) => lts.states[key].initial);
            if (!originalInitialStateKey) throw new Error("Initial state not found.");
            const initialStateKey = getRepresentativeState(originalInitialStateKey);

            currentlyVisibleMendedTransitions.push({
                source: initialStateKey,
                target: initialStateKey,
                label: "τ",
                weak: true,
                x: lts.states[initialStateKey].x - 60,
                y: lts.states[initialStateKey].y,
                transitionIndex: -1
            });
        }

        // methods

        const initializePartition = (): void => {
            partitioningInitialized = true;
            saveState();
        };

        const getPartitionCircles = (): ReactElement => <g>
            { mapObjectToArray(lts.states, (state, stateKey) => <circle key={stateKey} className="partition" r="28" style={{ transform: shiftToTransform(state), fill: colors[statePartitions[stateKey]]}} />) }
        </g>;

        const getPartitionCirclesForMinimizedLts = (): ReactElement => <g>
            { mapObjectToArray(representativeStates, (state, stateKey) => <circle key={stateKey} className="partition" r="28" style={{ transform: shiftToTransform(state), fill: colors[statePartitions[stateKey]]}} />) }
        </g>;

        const getHighlights = (): ReactElement => <g>
            { highlighted.map((stateKey) => <circle key={stateKey} className="highlight" r="32" style={{ transform: shiftToTransform(lts.states[stateKey]) }} />) }
        </g>;

        const getTransitionSourceHighlight = (): ReactElement | null => selectedTransitionSource ? <circle className="highlight" r="32" style={{ transform: shiftToTransform(lts.states[selectedTransitionSource]) }} /> : null; 

        const [showSplitMissedModal, setShowSplitMissedModal] = useState<boolean>(false);

        const selectSplitter = (action: string, partition: number): void => {
            currentSplitter = { action, partition };
            saveState();
        };

        const unselectSplitter = (): void => {
            setHighlighted([]);
            currentSplitter = null;
            saveState();
        };

        const [showSplitWrongModal, setShowSplitWrongModal] = useState<boolean>(false);

        const finalizeSplit = (): void => {
            if (!currentSplitter) throw new Error("Cannot finalized split without in-progress split.");
            const splitResult = splitResults[currentSplitter.partition][currentSplitter.action];

            const overHightlighted = highlighted.filter((s) => !splitResult.hitStates.includes(s)); // all highlights that are wrong

            // looking for missing highlights is more complicated as we allow the user to omit full partitions
            const missingHighlights = [];
            partitions.forEach((partition) => {
                const statesInPartition = partitionStates[partition];
                if (statesInPartition.every((s) => splitResult.hitStates.includes(s) && !highlighted.includes(s))) return; // full partition, omission is OK
                statesInPartition.forEach((s) => {
                    if (splitResult.hitStates.includes(s) && !highlighted.includes(s)) missingHighlights.push(s);
                });
            });

            if (overHightlighted.length === 0 && missingHighlights.length === 0) {
                // so the user actually selected correctly – now we need to actually adapt the partitioning

                const newPartitions: number[] = [];

                let newPartition = 0;

                partitions.forEach((partition) => { // note that while we may create new partitions within the loop, partitions is not updated – this happens in the next cycle
                    const statesThatMove = partitionStates[partition].filter((s) => highlighted.includes(s));
                    
                    if (statesThatMove.length === 0) return; // nothing to do
                    if (statesThatMove.length === partitionStates[partition].length) return; // everything would move, let's not do that

                    // look for new partition id
                    while (partitions.includes(newPartition) || newPartitions.includes(newPartition)) newPartition++;
                    newPartitions.push(newPartition);

                    statesThatMove.forEach((s) => statePartitions[s] = newPartition); // move!
                });

                setHighlighted([]); // unselect all
                currentSplitter = null; // stop split

                saveState();
            } else {
                setShowSplitWrongModal(true);
                userStudy.reportEvent({ type: "INTERACTIVE_REPORT_ERROR", interactiveType: "interactive", interactiveId: id });
            }
        };

        const finalizePartitioning = (): void => {
            if (flatten(mapObjectToArray(splitResults, (o) => mapObjectToArray(o, (res) => res))).every(({ generatesSplit: splitted }) => !splitted)) {
                partitioningFinalized = true;
                saveState();
            } else {
                setShowSplitMissedModal(true);
                userStudy.reportEvent({ type: "INTERACTIVE_REPORT_ERROR", interactiveType: "interactive", interactiveId: id });
            }
        };

        const [showTransitionMissedModal, setShowTransitionMissedModal] = useState<boolean>(false);

        const finalizeTransitionReconstruction = (): void => {
            const foundAllTransitions = lts.transitions.every((originalTransition) => selectedMendedTransitions.some((newTransition) => originalTransition.label === newTransition.label && newTransition.source === getRepresentativeState(originalTransition.source) && newTransition.target === getRepresentativeState(originalTransition.target)));
            if (foundAllTransitions) {
                reconstructionFinalized = true;
                if (!weak) markAsSolved();
                saveState();
            } else {
                setShowTransitionMissedModal(true);
                userStudy.reportEvent({ type: "INTERACTIVE_REPORT_ERROR", interactiveType: "interactive", interactiveId: id });
            }
        };

        const handleStateClick = (stateKey: string): void => {
            if (!currentSplitter) return;

            if (highlighted.includes(stateKey)) {
                setHighlighted(highlighted.filter((s) => s !== stateKey));
            } else {
                setHighlighted([...highlighted, stateKey]);
            }
        };

        const [showTransitionWrongModal, setShowTransitionWrongModal] = useState<boolean>(false);

        const handleDrawTransitionAttempt = (stateKey: string): void => {
            if (reconstructionFinalized) return;
            if (actionForDrawing) {
                if (selectedTransitionSource) {
                    // user attempts to add a transition

                    if (selectedMendedTransitions.some(({ source, target, label}) => label === actionForDrawing && source === selectedTransitionSource && target === stateKey)) {
                        // attempted duplicate insertion
                        setSelectedTransitionSource(null);
                        return;
                    }

                    // validate whether the new transition is actually correct
                    const candidateOriginalTransitions = lts.transitions.map((t, i) => ({ index: i, transition: t })).filter(({ transition : { source, target, label } }) => label === actionForDrawing && getRepresentativeState(source) === selectedTransitionSource && getRepresentativeState(target) === stateKey);
                    if (candidateOriginalTransitions.length > 0) {
                        // we agree to add the requested transition
                        
                        // but we need to identify which of the candidates is best, visually speaking
                        const scoredOriginalTransitions = candidateOriginalTransitions.map(({ index, transition }) => ({
                            index,
                            transition,
                            score: (transition.source === selectedTransitionSource ? 1 : 0) + (transition.target === stateKey ? 1 : 0)
                        })).sort((a, b) => b.score - a.score);

                        mendedTransitionSelections.push(scoredOriginalTransitions[0].index);
                        setSelectedTransitionSource(null);
                        saveState();
                    } else {
                        if (selectedTransitionSource === stateKey) {
                            // Maybe the user didn't try to add a self-loop, but just wanted to deselect?
                            // ignore it
                            setSelectedTransitionSource(null);
                        } else {
                            setShowTransitionWrongModal(true);
                            userStudy.reportEvent({ type: "INTERACTIVE_REPORT_ERROR", interactiveType: "interactive", interactiveId: id });
                            setSelectedTransitionSource(null);
                        }
                    }
                } else {
                    setSelectedTransitionSource(stateKey);
                }
            }
        };

        const [showWrongNecessityAnswerModal, setShowWrongNecessityAnswerModal] = useState<boolean>(false);
        
        const handleTransitionNecessityAnswer = (transitionIsNeeded: boolean): void => {
            const transitionUnderInvestigation = selectedMendedTransitions[mendedTransitionNecessityCheckPosition];
            
            // first, we need to compute the correct answer
            const getPureTauSuccessors = (from: string): Set<string> => {
                const res = new Set([from]);
                let prevSize = 0;
                while (prevSize < res.size) { // fixpoint iteration
                    prevSize = res.size;
                    res.forEach((s) => currentlyVisibleMendedTransitions.filter((t) => t.weak && t.source === s).forEach(({ target }) => res.add(target)));
                }
                return res;
            };
            
            // try to emulate the transition
            let currentPositions = new Set([transitionUnderInvestigation.source]);
            currentPositions = union([...currentPositions].map((s) => getPureTauSuccessors(s)));
            if (!transitionUnderInvestigation.weak) {
                // non-tau transition: we still need to make the main step
                currentPositions = new Set([...currentPositions].flatMap((s) => currentlyVisibleMendedTransitions.filter(({ source, label }) => source === s && label === transitionUnderInvestigation.label).map(({ target }) => target)));
                // and another tau step
                currentPositions = union([...currentPositions].map((s) => getPureTauSuccessors(s)));
            }
            
            const transitionIsActuallyNeeded = ![...currentPositions].includes(transitionUnderInvestigation.target);
            
            if (transitionIsNeeded === transitionIsActuallyNeeded) {
                if (transitionIsNeeded) {
                    mendedTransitionNecessityCheckPosition++;
                } else {
                    mendedTransitionSelections.splice(mendedTransitionNecessityCheckPosition, 1);
                    // no need to increment mendedTransitionNecessityCheckPosition, it now points to the next transition
                }
                if (mendedTransitionNecessityCheckPosition === mendedTransitionSelections.length) {
                    transitionCleanupFinalized = true;
                    if (!observationCongruent) markAsSolved();
                }
                saveState();
            } else {
                setShowWrongNecessityAnswerModal(true);
                userStudy.reportEvent({ type: "INTERACTIVE_REPORT_ERROR", interactiveType: "interactive", interactiveId: id });
            }
        };

        const [showWrongSelfLoopAnswerModal, setShowWrongSelfLoopAnswerModal] = useState<boolean>(false);
        
        const handleTauSelfLoopInsertion = (selfLoopNeeded: boolean): void => {
            const selfLoopActuallyNeeded = lts.transitions.some(({ source, weak, target }) => lts.states[source].initial && weak && partitionStates[statePartitions[target]].some((s) => lts.states[s].initial)); // TODO fix this
            
            if (selfLoopNeeded === selfLoopActuallyNeeded) {
                addTauLoop = selfLoopNeeded;
                tauLoopInsertionFinalized = true;
                markAsSolved();
            } else {
                setShowWrongSelfLoopAnswerModal(true);
                userStudy.reportEvent({ type: "INTERACTIVE_REPORT_ERROR", interactiveType: "interactive", interactiveId: id });
            }
            saveState();
        };

        const { stage, progress } = partitioningFinalized ? (reconstructionFinalized ? (transitionCleanupFinalized ? { stage: 3, progress: 0 } : { stage: 2, progress: mendedTransitionNecessityCheckPosition }) : { stage: 1, progress: selectedMendedTransitions.length }) : { stage: 0, progress: partitions.length - 1 } ;

        useEffect(() => {
            userStudy.reportEvent({ type: "INTERACTIVE_PROGRESS_UPDATE", interactiveType: "interactive", interactiveId: id, stage, progress });
        }, [id, progress, stage, userStudy]);

        return <div className="minimization">
            <Modal className="minimization-modal" centered show={showSplitWrongModal} onHide={(): void => setShowSplitWrongModal(false)}>
                <Modal.Header closeButton>
                    <Modal.Title>Oh no!</Modal.Title>
                </Modal.Header>
                <Modal.Body>
                    <p>
                        <b>That selection is not correct.</b>
                    </p>
                    <p>
                        Double-check that you selected every state that has { weak ? <Fragment>a weak</Fragment> : <Fragment>an</Fragment> } outgoing <MathSpan content={currentSplitter?.action ?? ""} />-transition to a <span className="partition-icon" style={{ backgroundColor: colors[currentSplitter?.partition ?? 0] }}></span> state, and nothing else.
                    </p>
                    <p>
                        Make sure you always consider the whole LTS, not just transitions originating from a single partition.
                    </p>
                    <div className="subtle-notice">
                        You may omit partitions where <i>all</i> states have such an outgoing transition, but if you do make sure you do not select any state from that partition.
                    </div>
                </Modal.Body>
                <Modal.Footer>
                    <Button variant="primary" onClick={(): void => setShowSplitWrongModal(false)}>
                        OK
                    </Button>
                </Modal.Footer>
            </Modal>

            <Modal className="minimization-modal" centered show={showSplitMissedModal} onHide={(): void => setShowSplitMissedModal(false)}>
                <Modal.Header closeButton>
                    <Modal.Title>Oh no!</Modal.Title>
                </Modal.Header>
                <Modal.Body>
                    <p>
                        <b>It looks like you missed a split.</b>
                    </p>
                    <p>
                        If you do not find every possible way to split a partition, the result will be incorrect – it will not be bisimilar the the original LTS!
                    </p>
                    <p>
                        Carefully re-check every possible combination of partition and action to see whether the predecessor operation hits a partition partially.
                    </p>
                </Modal.Body>
                <Modal.Footer>
                    <Button variant="primary" onClick={(): void => setShowSplitMissedModal(false)}>
                        OK
                    </Button>
                </Modal.Footer>
            </Modal>

            <Modal className="minimization-modal" centered show={showTransitionMissedModal} onHide={(): void => setShowTransitionMissedModal(false)}>
                <Modal.Header closeButton>
                    <Modal.Title>Oh no!</Modal.Title>
                </Modal.Header>
                <Modal.Body>
                    <p>
                        <b>It looks like you missed a transition.</b>
                    </p>
                    <p>
                        Carefully go through all transitions in the original LTS and make sure there is a corresponding transition in the new LTS.
                        { weak ? <Fragment>Make sure not to miss any τ-transitions!</Fragment> : null }
                    </p>
                </Modal.Body>
                <Modal.Footer>
                    <Button variant="primary" onClick={(): void => setShowTransitionMissedModal(false)}>
                        OK
                    </Button>
                </Modal.Footer>
            </Modal>

            <Modal className="minimization-modal" centered show={showTransitionWrongModal} onHide={(): void => setShowTransitionWrongModal(false)}>
                <Modal.Header closeButton>
                    <Modal.Title>Oh no!</Modal.Title>
                </Modal.Header>
                <Modal.Body>
                    <p>
                        <b>That transition isn't right.</b>
                    </p>
                    <p>
                        Make sure you have selected the action which you are currently working on.
                        Then, only draw transitions between states where the corresponding original partitions also have a corresponding transition.
                    </p>
                </Modal.Body>
                <Modal.Footer>
                    <Button variant="primary" onClick={(): void => setShowTransitionWrongModal(false)}>
                        OK
                    </Button>
                </Modal.Footer>
            </Modal>

            <Modal className="minimization-modal" centered show={showWrongNecessityAnswerModal} onHide={(): void => setShowWrongNecessityAnswerModal(false)}>
                <Modal.Header closeButton>
                    <Modal.Title>Oh no!</Modal.Title>
                </Modal.Header>
                <Modal.Body>
                    <p>
                        <b>That answer isn't right.</b>
                    </p>
                    <p>
                        While the correct answer is obvious now, take the time to understand why.
                        Look for ways to piece together the missing transition.
                        For τ-transitions, use zero or more τ-transitions.
                        For other transitions, use zero or more τ-transitions, followed by exactly one transition with the right action, followed by zero or more τ-transitions.
                    </p>
                </Modal.Body>
                <Modal.Footer>
                    <Button variant="primary" onClick={(): void => setShowWrongNecessityAnswerModal(false)}>
                        OK
                    </Button>
                </Modal.Footer>
            </Modal>

            <Modal className="minimization-modal" centered show={showWrongSelfLoopAnswerModal} onHide={(): void => setShowWrongSelfLoopAnswerModal(false)}>
                <Modal.Header closeButton>
                    <Modal.Title>Oh no!</Modal.Title>
                </Modal.Header>
                <Modal.Body>
                    <p>
                        <b>That answer isn't right.</b>
                    </p>
                    <p>
                        While the correct answer is obvious know, take the time to understand why.
                        Look for a τ-transition from the original initial state to a state within the same partition.
                    </p>
                </Modal.Body>
                <Modal.Footer>
                    <Button variant="primary" onClick={(): void => setShowWrongSelfLoopAnswerModal(false)}>
                        OK
                    </Button>
                </Modal.Footer>
            </Modal>

            <CenterOrShrinkGraphics width={measurements.x}>
                <LTSSVGFrame darkMode={theme === 'dark'} hideWeak={false} largeFont={false} moreSvgProps={{ viewBox: `0 0 ${imageSize.x} ${imageSize.y}` }}>
                    <g style={{ transform: shiftToTransform({ x: imageMargin, y: imageMargin })}}>
                        { partitioningInitialized ? getPartitionCircles() : null }
                        { getHighlights() }
                        <LTSSVGFragmentView states={lts.states} transitions={lts.transitions} showExpandNotice={false} onStateClick={handleStateClick} />
                    </g>
                </LTSSVGFrame>
            </CenterOrShrinkGraphics>
            <div>
                { partitioningInitialized ? <Fragment>
                    { partitioningFinalized ? <Fragment>
                        <p>We've found the final partitioning. Now, we have to construct the minimized LTS.</p>
                        <CenterOrShrinkGraphics width={measurements.x}>
                            <LTSSVGFrame darkMode={theme === 'dark'} hideWeak={false} largeFont={false} moreSvgProps={{ viewBox: `0 0 ${imageSize.x} ${imageSize.y}` }}>
                                <g style={{ transform: shiftToTransform({ x: imageMargin, y: imageMargin })}}>
                                    { getPartitionCirclesForMinimizedLts() }
                                    { getTransitionSourceHighlight() }
                                    <LTSSVGFragmentView states={representativeStates} transitions={currentlyVisibleMendedTransitions} showExpandNotice={false} onStateClick={handleDrawTransitionAttempt} />
                                </g>
                            </LTSSVGFrame>
                        </CenterOrShrinkGraphics>
                        { reconstructionFinalized ? <Fragment>
                            { weak ? <Fragment>
                                { transitionCleanupFinalized ? <Fragment>
                                    { observationCongruent ? <Fragment>
                                        { tauLoopInsertionFinalized ? <Fragment>
                                            <p>All done!</p>
                                        </Fragment> : <Fragment>
                                            <p>One last step. To ensure the result is observation congruent, we need to check whether the original LTS has a τ-transition from the initial state to a state within the same partition.</p>
                                            <p>
                                                Is this the case?
                                                <Button className="action-button" size="sm" variant="primary" onClick={(): void => handleTauSelfLoopInsertion(true)}>Yes, add τ-self-loop</Button>
                                                <Button className="action-button" size="sm" variant="primary" onClick={(): void => handleTauSelfLoopInsertion(false)}>No, all done</Button>
                                            </p>
                                        </Fragment> }
                                    </Fragment> : <Fragment>
                                        <p>All done!</p>
                                    </Fragment> }
                                </Fragment> : <Fragment>
                                    <p>Now, we have to test whether each transition is actually required.</p>
                                    <p>
                                        We're currently investigating the transition <MathSpan content={`${selectedMendedTransitions[mendedTransitionNecessityCheckPosition].source} \\xrightarrow{${selectedMendedTransitions[mendedTransitionNecessityCheckPosition].label}} ${selectedMendedTransitions[mendedTransitionNecessityCheckPosition].target}`} />. It's temporarily hidden in the LTS above. Does <MathSpan content={`${selectedMendedTransitions[mendedTransitionNecessityCheckPosition].source} \\xRightarrow{${selectedMendedTransitions[mendedTransitionNecessityCheckPosition].label}} ${selectedMendedTransitions[mendedTransitionNecessityCheckPosition].target}`} /> still hold despite the temporary removal?
                                        <Button className="action-button" size="sm" variant="primary" onClick={(): void => handleTransitionNecessityAnswer(false)}>Yes, remove it</Button>
                                        <Button className="action-button" size="sm" variant="primary" onClick={(): void => handleTransitionNecessityAnswer(true)}>No, restore it</Button>
                                    </p>
                                </Fragment> }
                            </Fragment> : <Fragment>
                                <p>All done!</p>
                            </Fragment> }
                        </Fragment> : <Fragment>
                            <p>Above, we've selected one state per partition to represent the whole partition. To add the transitions, select an action, then add transitions between states whenever the original LTS had such a transition between any states of the corresponding partitions.</p>
                            <div>
                                Select an action <ButtonGroup size="sm">
                                    { actions.map((action, i) => <Button key={i} active={actionForDrawing === action} onClick={(): void => setActionForDrawing(action)}>{ action }</Button>) }
                                </ButtonGroup>, then tap or click source and target to draw transitions.
                            </div>
                            <p>
                                Once all transitions are added, select <Button className="action-button" size="sm" onClick={finalizeTransitionReconstruction}>Done</Button>
                            </p>
                        </Fragment> }
                    </Fragment> : <Fragment>
                        { currentSplitter ? <Fragment>
                            <p>We're now splitting from partition <span className="partition-icon" style={{ backgroundColor: colors[currentSplitter.partition] }}></span> with action <MathSpan content={currentSplitter.action} />. <Button size="sm" variant="warning" onClick={unselectSplitter}>Cancel</Button></p>
                            <p>Please click or tap all states that have { weak ? <Fragment>a weak</Fragment> : <Fragment>an</Fragment> } outgoing <MathSpan content={currentSplitter.action} />-transition to a <span className="partition-icon" style={{ backgroundColor: colors[currentSplitter.partition] }}></span> state. <Button className="action-button" onClick={finalizeSplit}>Done</Button></p>
                            <p className="hint">You may omit partitions where <i>all</i> states have such an outgoing transition.</p>
                        </Fragment> : <Fragment>
                            <p>Until the partitioning is finalized, we are looking for ways to split partitions. Can you find an action <MathSpan content="\alpha" /> and partition <MathSpan content="s" /> such that { weak ? <Fragment>
                                for at least one partition, only some of its states have a weak transition with <MathSpan content="\alpha" /> to a state in <MathSpan content="s" />?
                            </Fragment> : <Fragment>
                                <MathSpan content="\mathit{Pre}(s, \alpha)" /> hits some partitions partially?
                            </Fragment> } </p>
                            <ul>
                                { partitions.map((partition, partitionIndex) => actions.map((action, actionIndex) => <li key={`${partitionIndex}/${actionIndex}`}>
                                    <Button size="sm" onClick={(): void => selectSplitter(action, partition)}>Yes: Action <MathSpan content={action} />, Partition <span className="partition-icon" style={{ backgroundColor: colors[partition] }}></span></Button>
                                </li>)) }
                                <li><Button size="sm" onClick={finalizePartitioning}>No</Button></li>
                            </ul>
                        </Fragment>}
                    </Fragment> }
                    <p className="reset">
                        Want to start over?
                        <Button size="sm" variant="danger" className="action-button" onClick={resetState}>Reset Exercise</Button>
                    </p>
                </Fragment> : <Fragment>
                    <p>Our goal is to minimize the LTS above. The first step is to assign every state to the same partition. <Button className="action-button" onClick={initializePartition}>Do it</Button></p>
                </Fragment> }
            </div>
        </div>;
    };

    return Minimization;
};

export default MinimizationBuilder;
