import React, { useContext, Fragment, ReactElement, useState, useEffect } from 'react';
import { InteractiveComponent } from '../../../hybrid-documents-reader-core/interactive/InteractiveElement';
import { CurrentThemeContext } from '../../../hybrid-documents-reader-core/ThemeApplicator';
import LTSSVGFrame from '../../../pseuco-shared-components/ui/editors/lts/LTSSVGFrame';
import { LTSViewData } from '../../../pseuco-shared-components/ui/editors/lts/LTSInteractiveView';
import { Button, Modal } from 'react-bootstrap';
import { intersperse, intersperseWithMainLast } from '../../../pseuco-shared-components/util/array';
import { flatten } from '../../../hybrid-documents-reader-core/util/functional';

import "./provingBisimilarity.scss";

import { mutatingFilter } from '../../../pseuco-shared-components/util/arrayMutations';
import MathSpan from '../../../pseuco-shared-components/ui/mathSpan/MathSpan';
import { DoubleSided, Side, isSide, otherSide, shiftToCoords, addShifts, prepareDoubleLTSForRendering, Pair, shiftToTransform, sides } from '../DoubleSidedLTS';
import CenterOrShrinkGraphics from '../CenterOrShrinkGraphics';
import { mapObjectToArray } from '../../../pseuco-shared-components/util/functional';
import { LTSTransitionViewDataAlike, generateWeakTransitions } from '../ltsUtils';
import { UserStudyContext } from '../../study/UserStudy';
import classNames from 'classnames';

const bezierShift = { x: 0, y: -80 };

const ProvingBisimilarityBuilder = async (lts: DoubleSided<LTSViewData>, initialTuple: DoubleSided<string>, {
    showFirstTimeHints = false,
    solvedOnConflicts = null,
    suggestBuildingRelationFirst = false,
    weak = false,
    observationCongruent = false,
}: {
    showFirstTimeHints?: boolean;
    solvedOnConflicts?: null | DoubleSided<string>[];
    suggestBuildingRelationFirst?: boolean;
    weak?: boolean;
    observationCongruent?: boolean;
}): Promise<InteractiveComponent> => {
    if (observationCongruent && !weak) throw new Error("Inconsistent options for provingBisimilarity interactive element.");

    const innerMargin = 100;
    const margins = {
        top: 50, // to accommodate relation
        bottom: 20,
        left: 20,
        right: 20
    };
    
    const { doubleLtsElement, shifts, imageSize } = prepareDoubleLTSForRendering(lts, margins, innerMargin);
    
    const weakTransitions: DoubleSided<LTSTransitionViewDataAlike[]> = { left: [], right: [] };
    if (weak) {
        sides.forEach((side) => weakTransitions[side] = generateWeakTransitions(lts[side]));
    }

    const ProvingBisimilarity: InteractiveComponent = ({ state, setState, markAsSolved, id }) => {
        let initialized = false;

        const theme = useContext(CurrentThemeContext);
        const userStudy = useContext(UserStudyContext);

        let relation: Pair[] = [];
        if (state && Array.isArray(state.relation)) {
            // carefully copy over relation from (non-typesafe) state
            state.relation.forEach((relationEntry: Pair) => {
                if (typeof relationEntry.left === "string" && relationEntry.left in lts.left.states && typeof relationEntry.right === "string" && relationEntry.right in lts.right.states) relation.push(relationEntry);
            });
        }

        type DefenseSelection = {
            attackingSide: Side;
            attacker: string;
            attackerTarget: string;
            action: string;
            defender: string;
            defenderTarget: string;
        };
    
        let selectedDefenses: DefenseSelection[] = [];
        if (state && Array.isArray(state.selectedDefenses)) {
            // carefully copy over selections from (non-typesafe) state
            state.selectedDefenses.forEach((selectedDefense: DefenseSelection) => {
                if (
                    typeof selectedDefense.attacker === "string"
                    && typeof selectedDefense.attackerTarget === "string"
                    && typeof selectedDefense.action === "string"
                    && typeof selectedDefense.defender === "string"
                    && typeof selectedDefense.defenderTarget === "string"
                ) selectedDefenses.push(selectedDefense);
            });
        }

        let seenConflictsOn: DoubleSided<string>[] = [];
        if (state && Array.isArray(state.seenConflictsOn)) {
            // carefully copy over list from (non-typesafe) state
            state.seenConflictsOn.forEach((entry: Pair) => {
                if (typeof entry.left === "string" && typeof entry.right === "string") seenConflictsOn.push(entry);
            });
        }

        let relationAcknowledged = false;
        if (state && typeof state.relationAcknowledged === "boolean") relationAcknowledged = state.relationAcknowledged;
        
        let confirmedObservationCongruence = false;
        if (state && typeof state.confirmedObservationCongruence === "boolean") confirmedObservationCongruence = state.confirmedObservationCongruence;
    
        let shownRelationPairRestrictionsModal = false;
        if (state && typeof state.shownRelationPairRestrictionsModal === "boolean") shownRelationPairRestrictionsModal = state.shownRelationPairRestrictionsModal;

        const arrayHasPair = (array: DoubleSided<string>[], pair: DoubleSided<string>): boolean => array.some(({ left, right }) => left === pair.left && right === pair.right);

        const isValidState = (side: Side, stateKey: string): boolean => !!lts[side].states[stateKey];
    
        const hasTransition = (side: Side, from: string, action: string, to: string, weak: boolean): boolean => (weak ? weakTransitions[side] : lts[side].transitions).some((t) => t.source === from && t.label === action && t.target === to);
    
        const isPairInRelation = (pair: Pair): boolean => arrayHasPair(relation, pair);
    
        /**
         * Given an attacker and defender, and the information which side is attacking, creates the matching pair of the relation.
         * @param attackingSide The side that is attacking.
         * @param attacker The attacker's start (or target) state.
         * @param defender The defender's start (or target) state.
         */
        const makePairFromAttack = (attackingSide: Side, attacker: string, defender: string): Pair => ({ [attackingSide]: attacker, [otherSide(attackingSide)]: defender }) as Pair; // typecast: TS does not understand we are guaranteed to have both left and right
    
        const checkAndSaveState = (): void => {
            // check the relation only contains states from the LTS
            mutatingFilter(relation, (pair) => isValidState("left", pair.left) && isValidState("right", pair.right));
    
            // check the defense selections are valid and correspond to pairs that are in the relation
            mutatingFilter(selectedDefenses, ({ attackingSide, attacker, attackerTarget, action, defender, defenderTarget }) => 
                isSide(attackingSide)
                && isValidState(attackingSide, attacker)
                && isValidState(attackingSide, attackerTarget)
                && hasTransition(attackingSide, attacker, action, attackerTarget, false)
                && isValidState(otherSide(attackingSide), defender)
                && isValidState(otherSide(attackingSide), defenderTarget)
                && hasTransition(otherSide(attackingSide), defender, action, defenderTarget, weak)
                && isPairInRelation(makePairFromAttack(attackingSide, attacker, defender)) // tests whether the selection corresponds to a pair (instead of being unnecessary)
            );

            mutatingFilter(seenConflictsOn, (pair) => isValidState("left", pair.left) && isValidState("right", pair.right));

            if (initialized) {
                if (!observationCongruent || !allAttacksProperlyDefended) confirmedObservationCongruence = false;
            } else {
                // allAttacksProperlyDefended isn't actually defined yet, so we need to defer this check
            }
    
            setState({ relation, selectedDefenses, seenConflictsOn, relationAcknowledged, shownRelationPairRestrictionsModal, confirmedObservationCongruence });
        };
    
        const pushPair = (pair: Pair): void => { if (!isPairInRelation(pair)) {
            relation.push(pair);
            checkAndSaveState();
        } };
    
        const removePairAtIndex = (index: number): void => {
            relation.splice(index, 1);
            checkAndSaveState();
            userStudy.reportEvent({ type: 'INTERACTIVE_REPORT_ERROR', interactiveType: "interactive", interactiveId: id });
        };
        
        const addConflict = (collision: DoubleSided<string>): void => {
            if (!arrayHasPair(seenConflictsOn, collision)) {
                seenConflictsOn.push(collision);
                checkAndSaveState();
            }
        };

        /**
         * Selects a defense for s -α-> s' from t or clears the selection.
         * @param attacker s
         * @param attackerTarget s'
         * @param action α
         * @param defender t
         * @param defenderTarget the target with t -α-> t', or `null` to clear the selection
         */
        const selectDefense = (attackingSide: Side, attacker: string, attackerTarget: string, action: string, defender: string, defenderTarget: string | null): void => {
            // first, remove previous selections
            selectedDefenses = selectedDefenses.filter(({ attackingSide: aS, attacker: a, attackerTarget: aT, action: act, defender: d }) => !(aS === attackingSide && a === attacker && aT === attackerTarget && act === action && d === defender));
    
            if (defenderTarget !== null) {
                selectedDefenses.push({
                    attackingSide,
                    attacker,
                    attackerTarget,
                    action,
                    defender,
                    defenderTarget
                });
            }
            if (defenderTarget === null) {
                userStudy.reportEvent({ type: 'INTERACTIVE_REPORT_ERROR', interactiveType: "interactive", interactiveId: id });
            }
    
            checkAndSaveState();
        };
    
        const getSelectedDefense = (attackingSide: Side, attacker: string, attackerTarget: string, action: string, defender: string): string | null => {
            const hit = selectedDefenses.find(({ attackingSide: s, attacker: a, attackerTarget: aT, action: act, defender: d }) => (s === attackingSide && a === attacker && aT === attackerTarget && act === action && d === defender));
            if (!hit) return null;
            return hit.defenderTarget;
        };
    
        type Attack = {
            attackingSide: Side;
            attacker: string;
            attackerTarget: string;
            action: string;
            initialTau: boolean; // whether this attack is based on a τ transition and both sides are in their initial states
            defender: string;
            pair: Pair;
            pairIndex: number;
            selectedDefense: null | {
                target: string;
                pair: Pair;
                pairIsInRelation: boolean;
                isVirtualSelfLoop: boolean; // whether this is a τ self loop that is not based on a strong transition
            };
            defenseOptions: {
                defenderTarget: string;
                needsNewPair: boolean;
                targetPair: Pair;
            }[];
        };
    
        const getAttacks = (attackingSide: Side): Attack[] => flatten(relation.map((pair, pairIndex) => {
            const attacker = pair[attackingSide];
            const defendingSide = otherSide(attackingSide);
            const defender = pair[defendingSide];
            const matchingAttackTransitions = lts[attackingSide].transitions.filter((t) => t.source === attacker);
            return matchingAttackTransitions.map((t) => {
                const action = t.label;
                const initialTau = (t.weak && lts[attackingSide].states[attacker].initial && lts[defendingSide].states[defender].initial) ?? false;
                const attackerTarget = t.target;
                const selectedDefenseTarget = getSelectedDefense(attackingSide, attacker, attackerTarget, action, defender);
                const defenseTransitions = weak ? weakTransitions[defendingSide] : lts[defendingSide].transitions;
                const defenseOptions = defenseTransitions.filter(({ source, label }) => source === defender && label === action).map(({ target }) => {
                    const defenderTarget = target;
                    const targetPair = makePairFromAttack(attackingSide, attackerTarget, defenderTarget);
                    return {
                        defenderTarget,
                        needsNewPair: !isPairInRelation(targetPair),
                        targetPair
                    };
                });
                if (defenseOptions.length === 0) addConflict(pair);

                return {
                    attackingSide,
                    attacker,
                    action,
                    initialTau,
                    attackerTarget,
                    defender,
                    pair,
                    pairIndex,
                    selectedDefense: selectedDefenseTarget === null ? null : {
                        target: selectedDefenseTarget,
                        pair: makePairFromAttack(attackingSide, attackerTarget, selectedDefenseTarget),
                        pairIsInRelation: isPairInRelation(makePairFromAttack(attackingSide, attackerTarget, selectedDefenseTarget)),
                        isVirtualSelfLoop: selectedDefenseTarget === pair[defendingSide] && (t.weak ?? false) && !lts[defendingSide].transitions.some((tt) => tt.source == pair[defendingSide] && tt.target === selectedDefenseTarget && tt.weak)
                    },
                    defenseOptions
                };
            });
        }));
    
        const attacks: DoubleSided<Attack[]> = { left: getAttacks("left"), right: getAttacks("right") };
    
        const containsInitialTuple = isPairInRelation(initialTuple);
        const attackDefenseStatus = Object.values(attacks).flatMap((as) => as.map((a) => a.selectedDefense?.pairIsInRelation ?? false));
        const properlyDefendedAttacks = attackDefenseStatus.filter((b) => b).length;
        const allAttacksProperlyDefended = attackDefenseStatus.every((b) => b);
    
        const initialTauAttacks = Object.values(attacks).flatMap((as) => as.filter((a) => a.initialTau));
        const initialTauAttacksDefendedByVirtualSelfLoops = initialTauAttacks.filter((a) => a.selectedDefense?.isVirtualSelfLoop);

        const proofCompleted = containsInitialTuple && allAttacksProperlyDefended && (!observationCongruent || confirmedObservationCongruence);
        if (proofCompleted) markAsSolved();
        
        initialized = true;

        useEffect(() => {
            if (!solvedOnConflicts) {
                userStudy.reportEvent({ type: "INTERACTIVE_PROGRESS_UPDATE", interactiveType: "interactive", interactiveId: id, stage: 0, progress: containsInitialTuple ? properlyDefendedAttacks : 0 });
            }
        }, [containsInitialTuple, id, properlyDefendedAttacks, userStudy]);

        const foundConflictsCount = (solvedOnConflicts ?? []).filter((c) => arrayHasPair(seenConflictsOn, c)).length;
        const foundAllConflicts = solvedOnConflicts && foundConflictsCount >= solvedOnConflicts.length;
        if (foundAllConflicts) markAsSolved();

        useEffect(() => {
            if (solvedOnConflicts) {
                userStudy.reportEvent({ type: "INTERACTIVE_PROGRESS_UPDATE", interactiveType: "interactive", interactiveId: id, stage: 0, progress: foundConflictsCount });
            }
        }, [id, foundConflictsCount, userStudy]);

        const resetProof = (): void => {
            relation = [];
            selectedDefenses = [];
            seenConflictsOn = [];
            relationAcknowledged = false;
            checkAndSaveState();
        };

        const [selection, setSelection] = useState<null | { side: Side; stateKey: string }>(null);

        const [showRelationPairRestrictionsModal, setShowRelationPairRestrictionsModal] = useState<boolean>(false);

        const warnAboutRelationPairRestrictions = (): void => {
            if (!shownRelationPairRestrictionsModal) {
                setShowRelationPairRestrictionsModal(true);
                shownRelationPairRestrictionsModal = true;
                checkAndSaveState();
            }
        };

        const stateClick = (side: Side, stateKey: string): void => {
            if (selection && selection.side === side) {
                if (selection.stateKey === stateKey) {
                    // clicked again, de-select
                    setSelection(null);
                    warnAboutRelationPairRestrictions();
                } else {
                    // clicked on different state on same side
                    setSelection({ side, stateKey });
                    warnAboutRelationPairRestrictions();
                }
            } else if (selection && selection.side !== side) {
                // make connection
                pushPair({ [side]: stateKey, [selection.side]: selection.stateKey } as DoubleSided<string>);
                setSelection(null);
            } else {
                setSelection({ side, stateKey });
            }
        };

        const getStateClickTargets = (side: Side): ReactElement => <Fragment>
            { mapObjectToArray(lts[side].states, (state, stateKey) => <circle className="click-target" key={stateKey} r="30" onClick={(): void => stateClick(side, stateKey)} style={{ transform: shiftToTransform(addShifts(shifts[side], state))}} />) };
        </Fragment>;

        const getHighlight = (): ReactElement | null => selection ? <circle className="highlight" r="30" style={{ transform: shiftToTransform(addShifts(shifts[selection.side], lts[selection.side].states[selection.stateKey]))}} /> : null;

        const getInitialTauAttacksList = (): ReactElement => {
            const elements = intersperseWithMainLast<{ tex: boolean, content: string }>(initialTauAttacks.map((a) => ({ tex: true, content: `${a.attacker} \\xrightarrow{${a.action}} ${a.attackerTarget}` })), { tex: false, content: ", " }, { tex: false, content: ", and " }, { tex: false, content: " and " });
            return <>
                { elements.map((e, i) => e.tex ? <MathSpan key={i} content={e.content} /> : <span key={i}>{ e.content }</span>) }
            </>;
        };

        const defenderArrow = weak ? "\\xRightarrow" : "\\xrightarrow";

        const renderDefenseUI = (attackingSide: Side): React.ReactNode => {
            const currentAttacks = attacks[attackingSide];
    
            return <ul>
                { currentAttacks.map(({ attacker, action, attackerTarget, defender, pair, pairIndex, selectedDefense, defenseOptions, initialTau }, i) => <li key={i}>
                    <p className={classNames({ attack: true, highlighted: observationCongruent && allAttacksProperlyDefended && !confirmedObservationCongruence && initialTau })}>
                        {/* defense proof text */}
                        <MathSpan content={`(${pair.left}, ${pair.right}) \\in \\mathcal{R}`} /> and <MathSpan content={`${attacker} \\xrightarrow{${action}} ${attackerTarget}`} />{ selectedDefense ? <span>, but also <MathSpan content={`${defender} ${defenderArrow}{${action}} ${selectedDefense.target}`} />{ selectedDefense.pairIsInRelation ? <span> and <MathSpan content={`(${selectedDefense.pair.left}, ${selectedDefense.pair.right}) \\in \\mathcal{R}`} /></span> : null }</span> : null }.
                    </p>
                    { selectedDefense ? (selectedDefense.pairIsInRelation ? null :
                        // add target pair (or deselect defense) UI
                        <p>
                            <span className="instructions">The pair of the states we reached, <MathSpan content={`(${selectedDefense.pair.left}, ${selectedDefense.pair.right})`} />, is not in <MathSpan content="\mathcal{R}" />.</span>
                            <Button className="fix-it" onClick={(): void => pushPair(selectedDefense.pair)}>Add pair <MathSpan content={`(${selectedDefense.pair.left}, ${selectedDefense.pair.right})`} /> to <MathSpan content="\mathcal{R}" /></Button>
                            <Button className="fix-it" variant="secondary" onClick={(): void => selectDefense(attackingSide, attacker, attackerTarget, action, defender, null)}>Select different defending move</Button>
                        </p>
                    ) :
                        // defense selection UI
                        <p>
                            <span className="instructions">We need to simulate <MathSpan content={`${attacker} \\xrightarrow{${action}} ${attackerTarget}`} /> with a defending move <MathSpan content={`${defender} ${defenderArrow}{${action}} t'`} /> such that <MathSpan content={`(${attackerTarget}, t') \\in \\mathcal{R}`} />.</span>
                            { 
                                (defenseOptions.length > 0 ?
                                    // defending is possible, show selector
                                    <span>
                                        { defenseOptions.map(({ defenderTarget }, i) => <Button className="fix-it" key={i} onClick={(): void => selectDefense(attackingSide, attacker, attackerTarget, action, defender, defenderTarget)}>Select <MathSpan content={`${defender} ${defenderArrow}{${action}} ${defenderTarget}`} /></Button>) }
                                        <Button className="fix-it" variant="secondary" onClick={(): void => removePairAtIndex(pairIndex)}>Remove <MathSpan content={`(${pair.left}, ${pair.right})`} /> from <MathSpan content="\mathcal{R}" /></Button>
                                    </span>
                                    : <span className="oh-no"><br />No options to defend. Something is wrong! <Button className="fix-it" onClick={(): void => removePairAtIndex(pairIndex)}>Remove <MathSpan content={`(${pair.left}, ${pair.right})`} /> from <MathSpan content="\mathcal{R}" /></Button><br />Remember you can remove arbitrary pairs from <MathSpan content="\mathcal{R}" /> by clicking on them in the definition of <MathSpan content="\mathcal{R}" />.</span>
                                )
                            }
                        </p>
                    }
                </li>) }
            </ul>;
        };
    
        return <div className="proving-bisimilarity">
            <Modal centered show={showRelationPairRestrictionsModal} onHide={(): void => setShowRelationPairRestrictionsModal(false)}>
                <Modal.Header closeButton>
                    <Modal.Title>Could Not Add Pair to Relation</Modal.Title>
                </Modal.Header>
                <Modal.Body>
                    <p>
                        You clicked a state on the same “side” as the state that you previously selected.
                        This is formally valid, but does not help us, as the states we want to compare are on different sides.
                        No pair will be inserted, so you can use this to easily de-select a state or select a different one.
                    </p>
                    <div className="subtle-notice">
                        This message will only be shown once for this exercise.
                    </div>
                </Modal.Body>
                <Modal.Footer>
                    <Button variant="primary" onClick={(): void => setShowRelationPairRestrictionsModal(false)}>
                        OK
                    </Button>
                </Modal.Footer>
            </Modal>

            <div className="exercise-container">
                <div>
                    <p>In this exercise, we will consider the following LTS and { solvedOnConflicts ? <span>attempt to</span> : null } prove that states <MathSpan content={initialTuple.left} /> and <MathSpan content={initialTuple.right} /> are { observationCongruent ? "observation congruent" : weak ? "weakly bisimilar" : "bisimilar" }:</p>
                </div>
                <CenterOrShrinkGraphics width={imageSize.x}>
                    <LTSSVGFrame darkMode={theme === 'dark'} hideWeak={false} largeFont={false} moreSvgProps={{ viewBox: `0 0 ${imageSize.x} ${imageSize.y}` }}>
                        <g>
                            { getHighlight() }
                            {relation.map(({ left, right }, i) => <path
                                key={i}
                                className="relation-tuple"
                                d={`M ${shiftToCoords(addShifts(shifts.left, lts.left.states[left]))} C ${shiftToCoords(addShifts(shifts.left, lts.left.states[left], bezierShift))}, ${shiftToCoords(addShifts(shifts.right, lts.right.states[right], bezierShift))}, ${shiftToCoords(addShifts(shifts.right, lts.right.states[right]))}`}
                            ></path>)}
                        </g>
                        { doubleLtsElement }
                        <g>
                            { getStateClickTargets("left") }
                            { getStateClickTargets("right") }
                        </g>
                    </LTSSVGFrame>
                </CenterOrShrinkGraphics>
            </div>
            <div className="proof-container">
                { showFirstTimeHints ?
                    <Fragment>
                        <p>To prove <MathSpan content={initialTuple.left} /> and <MathSpan content={initialTuple.right} /> are bisimilar, we define a relation that contains <MathSpan content={`(${initialTuple.left}, ${initialTuple.right})`} /> and show that it is a bisimulation. Note that every pair we add to <MathSpan content="\mathcal{R}" /> generates new proof goals.</p>
                        <p>In this example, you can't really go wrong. Still, don't blindly click through, but carefully consider what you are doing – future examples will not be as easy!</p>
                    </Fragment>
                    : null }
                { solvedOnConflicts ? <p>
                    Since the states are actually <em>not</em> bisimilar, the proof will fail. Still, give it you best attempt, and try all options available until you run into a conflict. Try to observe what properties of the states in question prevent you from completing the proof.
                </p> : null}
                <p>Proof. Let <MathSpan content="\mathcal{R} \coloneqq \{" />{
                    intersperse(relation.map(({ left, right }, relationIndex) => (keyIndex: number): React.ReactElement => <MathSpan className="pair" key={keyIndex} content={`(${left}, ${right})`} onClick={(): void => removePairAtIndex(relationIndex)} />), (keyIndex) => <span key={keyIndex}>, </span>).map((e, i) => e(i)) // eslint-disable-line react/display-name
                }<MathSpan content="\}" />. {relation.length > 0 ? <span className="hint">Click a pair to remove it.</span> : null} <span className="hint">Add pairs by clicking the states in the LTS above or through the proof below.</span>
                </p>
                { (suggestBuildingRelationFirst && !relationAcknowledged) ? <div>
                    To finish the proof more effectively, it is often useful to attempt to find the relation first, before starting to consider the cases.
                    Add pairs to the relation by clicking the states you want to relate in the LTS above. When you're done, we'll continue the proof.
                    <Button className="fix-it" onClick={(): void => {
                        relationAcknowledged = true;
                        checkAndSaveState();
                    }}>Done</Button>
                </div> : <Fragment>
                    { containsInitialTuple ? <p>We trivially see <MathSpan content={`(${initialTuple.left}, ${initialTuple.right}) \\in \\mathcal{R}`} />, so it only remains to show that <MathSpan content="\mathcal{R}" /> is a { weak ? "weak bisimulation" : "bisimulation" }.</p> :
                        <p className="instructions">The relation does not contain the states we want to compare. <Button className="fix-it" onClick={(): void => pushPair(initialTuple)}>Add pair <MathSpan content={`(${initialTuple.left}, ${initialTuple.right})`} /> to <MathSpan content="\mathcal{R}" /></Button></p>
                    }
                    {
                        relation.length > 0 ? <div>
                            <p>We consider both directions separately.</p>
                            <ul>
                                <li>
                                    We first consider every pair in <MathSpan content="(s, t) \in \mathcal{R}" /> and every transition <MathSpan content="s \xrightarrow\alpha s'" />:
                                    {renderDefenseUI("left")}
                                </li>
                                <li>
                                    Now, we consider every pair in <MathSpan content="(s, t) \in \mathcal{R}" /> and every transition <MathSpan content="t \xrightarrow\alpha t'" />:
                                    {renderDefenseUI("right")}
                                </li>
                            </ul>
                            { allAttacksProperlyDefended ? <p>
                                All cases considered.
                                { observationCongruent ? <span>
                                    {" "}This proves that <MathSpan content={initialTuple.left} /> and <MathSpan content={initialTuple.right} /> are weakly bisimilar.
                                    {" "}{ confirmedObservationCongruence ? <span>
                                        Since in the proof above, all initial τ transitions (i.e. { getInitialTauAttacksList() }) were defended in accordance with the definition of observation congruence (i.e. not by “standing still”), <MathSpan content={initialTuple.left} /> and <MathSpan content={initialTuple.right} /> are also observation congruent.
                                    </span> : <span>
                                        <span className="instructions">To show that they are observation congruent as well, we need to check that every initial τ transition (i.e. { getInitialTauAttacksList() }, highlighted above) was defended using a real τ transition.</span>
                                        { initialTauAttacksDefendedByVirtualSelfLoops.length > 0 ? <span className="oh-no">
                                            {" "}This is not the case: The attack <MathSpan content={`${initialTauAttacksDefendedByVirtualSelfLoops[0].attacker} \\xrightarrow{${initialTauAttacksDefendedByVirtualSelfLoops[0].action}} ${initialTauAttacksDefendedByVirtualSelfLoops[0].attackerTarget}`} /> is defended by “standing still”.
                                            <Button className="fix-it" variant="primary" onClick={(): void => selectDefense(initialTauAttacksDefendedByVirtualSelfLoops[0].attackingSide, initialTauAttacksDefendedByVirtualSelfLoops[0].attacker, initialTauAttacksDefendedByVirtualSelfLoops[0].attackerTarget, initialTauAttacksDefendedByVirtualSelfLoops[0].action, initialTauAttacksDefendedByVirtualSelfLoops[0].defender, null)}>Select different defending move</Button>
                                        </span> : <span className="instructions">
                                            {" "}Please check this above, then press <Button className="fix-it" variant="primary" onClick={() => {
                                                confirmedObservationCongruence = true;
                                                checkAndSaveState();
                                            }}>Confirm</Button>
                                        </span> }
                                    </span> }
                                </span> : null }
                                { proofCompleted ? <span className="qed">■</span> : null }
                            </p> : null }
                        </div> : null
                    }
                </Fragment> }
                { proofCompleted ? <p>That's it! We've shown both LTS are { weak ? "weakly bisimilar" : "bisimilar" }. Want to try again? <Button className="fix-it" variant="danger" onClick={(): void => resetProof()}>Reset Proof</Button></p> : null }
                { foundAllConflicts ? <p>Good work! As you have seen, the proof is impossible to complete. Want to try again? <Button className="fix-it" variant="danger" onClick={(): void => resetProof()}>Reset Proof</Button></p> : null }
            </div>
        </div>;
    };

    return ProvingBisimilarity;
};

export default ProvingBisimilarityBuilder;
