import { FlowElement, flowReachabilityGraphValidator, FlowReachabilityGraphValidatorConfiguration } from "../../flowValidation";
import { moveValidatorDisallowDataRaces } from "../../moveSetValidators";
import { moveValidatorAnd, moveValidatorInitializeMutex, moveValidatorLockUnlockMutex, moveValidatorParticipatingAgents, moveValidatorPrintLn, moveValidatorReadPseuCoGlobal, moveValidatorStartAgent, moveValidatorWritePseuCoGlobal } from "../../moveValidators";
import { syntaxValidatorRestrictVariableUsage } from "../../syntaxValidation";
import { registerValidationWorkerCallback } from "../../validation";

const countDownAgentFlowElements = (agentId: number): FlowElement[] => {
    const singleStep = (step: number): FlowElement[] => [
        {
            moveCompletesStep: moveValidatorAnd(moveValidatorParticipatingAgents([agentId]), moveValidatorLockUnlockMutex("lock", 1)),
            description: `agent ${agentId}, iteration ${step}: lock the lock`,
            inlineDescription: `lock the lock (agent ${agentId}, iteration ${step})`
        },
        {
            moveCompletesStep: moveValidatorAnd(moveValidatorParticipatingAgents([agentId]), moveValidatorReadPseuCoGlobal("n", null, true)),
            description: `agent ${agentId}, iteration ${step}: read n`,
            inlineDescription: `read n (agent ${agentId}, iteration ${step})`
        },
        {
            moveCompletesStep: moveValidatorAnd(moveValidatorParticipatingAgents([agentId]), moveValidatorWritePseuCoGlobal("n", null, true)),
            description: `agent ${agentId}, iteration ${step}: write n`,
            inlineDescription: `write n (agent ${agentId}, iteration ${step})`
        },
        {
            moveCompletesStep: moveValidatorAnd(moveValidatorParticipatingAgents([agentId]), moveValidatorLockUnlockMutex("unlock", 1)),
            description: `agent ${agentId}, iteration ${step}: unlock the lock`,
            inlineDescription: `unlock the lock (agent ${agentId}, iteration ${step})`
        }
    ];

    return [...singleStep(1), ...singleStep(2), ...singleStep(3)];
};

const validatorConfiguration: FlowReachabilityGraphValidatorConfiguration = {
    steps: {
        order: "sequential",
        steps: [
            {
                order: "parallel",
                steps: [
                    {
                        moveCompletesStep: moveValidatorWritePseuCoGlobal("n", 6, false),
                        description: "initialize n to value 6"
                    },
                    {
                        moveCompletesStep: moveValidatorInitializeMutex(),
                        description: "initialized a lock"
                    }
                ],
            },
            {
                moveCompletesStep: moveValidatorAnd(moveValidatorParticipatingAgents([0, 1]), moveValidatorStartAgent()),
                description: "start a countDown agent"
            },
            {
                order: "parallel",
                steps: [
                    {
                        order: "sequential",
                        steps: countDownAgentFlowElements(1)
                    },
                    {
                        order: "sequential",
                        steps: [
                            {
                                moveCompletesStep: moveValidatorAnd(moveValidatorParticipatingAgents([0, 2]), moveValidatorStartAgent()),
                                description: "start a second countDown agent"
                            },
                            ...countDownAgentFlowElements(2)
                        ]
                    }
                ]
            },
            {
                moveCompletesStep: moveValidatorAnd(moveValidatorParticipatingAgents([0]), moveValidatorReadPseuCoGlobal("n", 0, false)),
                description: `read n = 0 from the main agent`
            },
            {
                moveCompletesStep: moveValidatorAnd(moveValidatorParticipatingAgents([0]), moveValidatorPrintLn(/^The final value is: 0$/, false)),
                description: `print "The final value is: 0" from the main agent`
            }
        ]
    },
    restrictedMoves: [
        {
            detector: moveValidatorReadPseuCoGlobal("n", null, true),
            description: "read n"
        },
        {
            detector: moveValidatorWritePseuCoGlobal("n", null, true),
            description: "write n"
        }
    ],
    isMoveSetAllowed: moveValidatorDisallowDataRaces()
};

const syntaxChecks = [
    syntaxValidatorRestrictVariableUsage("n", { assign: 1, postfix: 1, identifier: 1 })
];

registerValidationWorkerCallback(flowReachabilityGraphValidator(validatorConfiguration), { mp: false, sm: true }, syntaxChecks);
