import { Monitor, PrimitiveStatement, Struct } from "@pseuco/lang";
import { flowReachabilityGraphValidator, FlowReachabilityGraphValidatorConfiguration } from "../../flowValidation";
import { moveValidatorDisallowDataRaces } from "../../moveSetValidators";
import { moveValidatorAnd, moveValidatorInitializeMutex, moveValidatorLockUnlockMutex, moveValidatorParticipatingAgents, moveValidatorPrintLn, moveValidatorReadPseuCoStruct, moveValidatorStartAgent, moveValidatorWritePseuCoStruct } from "../../moveValidators";
import { SyntaxValidator, syntaxValidatorAllowNodeOnlyWithinNode, syntaxValidatorBanNode } from "../../syntaxValidation";
import { registerValidationWorkerCallback } from "../../validation";

let foundLoop = false;

const validatorConfiguration: FlowReachabilityGraphValidatorConfiguration = {
    steps: {
        order: "sequential",
        steps: [
            {
                moveCompletesStep: moveValidatorInitializeMutex(),
                description: "initialize a lock"
            },
            {
                moveCompletesStep: moveValidatorAnd(moveValidatorParticipatingAgents([0, 1]), moveValidatorStartAgent()),
                description: `start agent 1`
            },
            {
                order: "parallel",
                steps: [
                    {
                        order: "sequential", 
                        steps: [ // printer
                            {
                                moveCompletesStep: moveValidatorAnd(moveValidatorParticipatingAgents([1]), moveValidatorLockUnlockMutex("lock", 1, false)),
                                description: "agent 1: lock the lock"
                            },
                            {
                                moveCompletesStep: moveValidatorAnd(moveValidatorParticipatingAgents([1]), moveValidatorReadPseuCoStruct(0, "message", 42, true)),
                                description: "agent 1: read 42 from the message field of structure 0"
                            },
                            {
                                moveCompletesStep: moveValidatorAnd(moveValidatorParticipatingAgents([1]), moveValidatorLockUnlockMutex("unlock", 1, false)),
                                description: "agent 1: unlock the lock"
                            },
                            {
                                moveCompletesStep: moveValidatorAnd(moveValidatorParticipatingAgents([1]), moveValidatorPrintLn(/^42$/, false)),
                                description: "agent 1: print 42"
                            }
                        ]
                    },
                    {
                        order: "sequential",
                        steps: [
                            {
                                moveCompletesStep: moveValidatorAnd(moveValidatorParticipatingAgents([0, 2]), moveValidatorStartAgent()),
                                description: `start agent 2`
                            },
                            {
                                order: "sequential",
                                steps: [ // putter
                                    {
                                        moveCompletesStep: moveValidatorAnd(moveValidatorParticipatingAgents([2]), moveValidatorLockUnlockMutex("lock", 1, false)),
                                        description: "agent 2: lock the lock"
                                    },
                                    {
                                        moveCompletesStep: moveValidatorAnd(moveValidatorParticipatingAgents([2]), moveValidatorWritePseuCoStruct(0, "message", 42, false)),
                                        description: "agent 2: write 42 to the message field of structure 0"
                                    },
                                    {
                                        moveCompletesStep: moveValidatorAnd(moveValidatorParticipatingAgents([2]), moveValidatorLockUnlockMutex("unlock", 1, false)),
                                        description: "agent 2: unlock the lock"
                                    },
                                ]
                            }
                        ]
                    }
                ]
            }
        ]
    },
    isMoveSetAllowed: moveValidatorDisallowDataRaces(),
    restrictedMoves: [
        {
            detector: moveValidatorInitializeMutex(),
            description: "initialized a lock"
        }
    ],
    allowCycleOnMoves: () => {
        foundLoop = true;
        return { success: true, value: null };
    },
    endSearchAllowed: () => foundLoop ? { success: true, value: null } : { success: false, error: "Your program does not allow agent 1 to read -1 infinitely often." }
};

const syntaxChecks: SyntaxValidator[] = [
    syntaxValidatorAllowNodeOnlyWithinNode((node) => node instanceof PrimitiveStatement && (node.primitive === "lock" || node.primitive === "unlock"), (node) => node instanceof Struct, "In this exercise, locks may only be used from within the struct."),
    syntaxValidatorBanNode((node) => node instanceof Monitor, "You used a monitor. Good thought, but the monitor keyword will only be introduced later on. In this exercise, use a manually declared lock instead, please.")
];

registerValidationWorkerCallback(flowReachabilityGraphValidator(validatorConfiguration), { mp: false, sm: true }, syntaxChecks);
