import { Move } from "@pseuco/colored-petri-nets";
import { ValidationChecklistState } from "../../../pseuco-shared-components/ui/debugger/PseuCoDebugger";
import { MoveValidator } from "./moveValidators";
import { PartValidationResult, ReachabilityGraphValidator } from "./validation";


/**
 * A validator for a single step, plus a description.
 */
export type DescribedStep = {
    moveCompletesStep: MoveValidator;
    description: string;

    /**
     * An abbreviated description that is used when inserted into a sentence like "the program was expected to [.]"
     */
    inlineDescription?: string;
};


/**
 * Describes a single step, a sequence of steps, or a set of steps (to be performed in any order).
 */
export type FlowElement = DescribedStep | {
    order: "sequential",
    steps: FlowElement[]
} | {
    order: "parallel",
    steps: FlowElement[]
}


/**
 * Describes a simpler version of a [[ReachabilityGraphValidator]] that expects the program to perform a sequence of actions in a semi-restricted order.
 * Each step is described by a [[MoveValidator]].
 * The order of the steps can be enforced (sequential) or free (parallel), or any combination thereof (e.g. two sequentially ordered lists of actions can be allowed to run in parallel).
 * An additional function `isLegalMove` can be set that may inspect every move, allowing the validator to more easily ban certain actions globally.
 * Similarly, `isMoveSetAllowed` can provide a global check of all move _sets_, enabling the configuration to block certain patterns of concurrent behaviour (e.g. data races).
 * Finally, `restrictedMoves` allows restricting certain moves: A move that returns a truthy value here is only allowed if it progresses the flow.
 */
export type FlowReachabilityGraphValidatorConfiguration = {
    steps: FlowElement,
    isLegalMove?: (move: Move) => PartValidationResult<null>;
    isMoveSetAllowed?: (moves: Move[]) => PartValidationResult<null>;
    restrictedMoves?: {
        detector: (move: Move) => PartValidationResult<boolean>,
        description: string
    }[];
    allowCycleOnMoves?: (move: Move) => PartValidationResult<null>;
    endSearchAllowed?: () => PartValidationResult<null>;
}


/**
 * Remembers where execution within a [[FlowReachabilityGraphValidatorConfiguration]] is.
 */
type FlowExecutionState = {
    type: "single",
    state: boolean;
} | {
    type: "sequential",
    current: number, // next step that needs to be done (or last one, with a completed state, when all done)
    state: FlowExecutionState
} | {
    type: "parallel",
    states: FlowExecutionState[]
}


const flowGetInitialState = (element: FlowElement): FlowExecutionState => {
    if ("moveCompletesStep" in element) return {
        type: "single",
        state: false
    };
    switch (element.order) {
        case "sequential": {
            if (!element.steps.length) throw new Error("Sequential flow elements must have at least one step.");
            return {
                type: "sequential",
                current: 0,
                state: flowGetInitialState(element.steps[0])
            };
        }
        case "parallel": {
            return {
                type: "parallel",
                states: element.steps.map((s) => flowGetInitialState(s))
            };
        }
    }
};

class FlowStateInconsistencyError extends Error {
    constructor() {
        super("Internal inconsistency: The current flow execution state does not match the flow definition.");
    }
}

const flowIsCompleted = (element: FlowElement, state: FlowExecutionState): boolean => {
    if ("moveCompletesStep" in element) {
        if (state.type !== "single") throw new FlowStateInconsistencyError();
        return state.state;
    }
    switch (element.order) {
        case "sequential": {
            if (state.type !== "sequential") throw new FlowStateInconsistencyError();
            return state.current === element.steps.length - 1 && flowIsCompleted(element.steps[state.current], state.state);
        }
        case "parallel": {
            if (state.type !== "parallel") throw new FlowStateInconsistencyError();
            if (state.states.length !== element.steps.length) throw new FlowStateInconsistencyError();
            return element.steps.every((step, i) => flowIsCompleted(step, state.states[i]));
        }
    }
};

const flowAcceptMove = (element: FlowElement, state: FlowExecutionState, move: Move): PartValidationResult<FlowExecutionState> => {
    if (flowIsCompleted(element, state)) return { success: true, value: state }; // already completed, no change needed

    if ("moveCompletesStep" in element) {
        if (state.type !== "single") throw new FlowStateInconsistencyError();
        const moveValidationResult = element.moveCompletesStep(move);
        if (!moveValidationResult.success) return moveValidationResult;
        return { success: true, value: { type: "single", state: moveValidationResult.value } };
    }
    switch (element.order) {
        case "sequential": {
            if (state.type !== "sequential") throw new FlowStateInconsistencyError();
            const currentElement = element.steps[state.current];
            const moveValidationResult = flowAcceptMove(currentElement, state.state, move);
            if (!moveValidationResult.success) return moveValidationResult;
            if (flowIsCompleted(currentElement, moveValidationResult.value) && state.current < element.steps.length - 1) {
                // non-last step, advance
                return {
                    success: true,
                    value: {
                        type: "sequential",
                        current: state.current + 1,
                        state: flowGetInitialState(element.steps[state.current + 1])
                    }
                };
            } else {
                // last step, do not advance
                return {
                    success: true,
                    value: {
                        type: "sequential",
                        current: state.current,
                        state: moveValidationResult.value
                    }
                };
            }
        }
        case "parallel": {
            if (state.type !== "parallel") throw new FlowStateInconsistencyError();
            const newState: FlowExecutionState[] = [];

            for (let i = 0; i < element.steps.length; i++) {
                const subElement = element.steps[i];
                const subState = state.states[i];
                if (flowIsCompleted(subElement, subState)) {
                    // nothing to do here
                    newState.push(subState);
                    continue;
                }

                const moveValidationResult = flowAcceptMove(subElement, subState, move);
                if (!moveValidationResult.success) return moveValidationResult; // a parallel step was fully rejected, return rejection

                newState.push(moveValidationResult.value);
            }

            return {
                success: true,
                value: {
                    type: "parallel",
                    states: newState
                }
            };
        }
    }
};

const flowDescribeNextStep = (element: FlowElement, state: FlowExecutionState): string => {
    if (flowIsCompleted(element, state)) throw new Error("Internal inconsistency: Cannot describe next step of a completed state.");

    if ("moveCompletesStep" in element) {
        return element.inlineDescription ?? element.description;
    }
    switch (element.order) {
        case "sequential":
            if (state.type !== "sequential") throw new FlowStateInconsistencyError();
            return flowDescribeNextStep(element.steps[state.current], state.state);
        case "parallel": {
            if (state.type !== "parallel") throw new FlowStateInconsistencyError();
            const possibleSteps = element.steps.flatMap((subElement, i) => {
                if (flowIsCompleted(subElement, state.states[i])) return [];
                return [flowDescribeNextStep(subElement, state.states[i])];
            });
            return possibleSteps.map((message, i) => i === possibleSteps.length - 1 ? message : i === possibleSteps.length - 2 ? `${message}, or ` : `${message}, `).join("");
        }
    }
};

/**
 * Determines the validation checklist for a specific state.
 * @param element The flow element to get the validation checklist for.
 * @param state The state to get the validation checklist for. This supports two special values:
 *      - `true` indicates the element is already fully completed, so every entry should be given the `"DONE"` state.
 *      - `false` indicated the element is not active yet, so every entry should be given the `null` state.
 * @returns 
 */
const flowGetValidationChecklistState = (element: FlowElement, state: FlowExecutionState | boolean): ValidationChecklistState => {
    if ("moveCompletesStep" in element) {
        return {
            stages: [{
                description: element.description,
                status: state === false ? null : state === true ? "DONE" : flowIsCompleted(element, state) ? "DONE" : "WAITING"
            }]
        };
    }

    const stepValidationLists = element.steps.map((subElement, i) => {
        let subState: FlowExecutionState | boolean;
        if (typeof state === "boolean") {
            subState = state; // pass special states to subelement
        } else {
            switch (element.order) {
                case "sequential": {
                    if (state.type !== "sequential") throw new FlowStateInconsistencyError();
                    subState = state.current < i ? false : state.current === i ? state.state : true;
                    break;
                }
                case "parallel": {
                    if (state.type !== "parallel") throw new FlowStateInconsistencyError();
                    subState = state.states[i];
                    break;
                }
            }
        }
        return flowGetValidationChecklistState(subElement, subState);
    });

    return {
        stages: stepValidationLists.flatMap((vl) => vl.stages)
    };
};


/**
 * Turns a [[flowReachabilityGraphValidatorConfiguration]] into an actual [[ReachabilityGraphValidator]].
 */
export const flowReachabilityGraphValidator = (configuration: FlowReachabilityGraphValidatorConfiguration): ReachabilityGraphValidator<FlowExecutionState> => ({
    initialState: flowGetInitialState(configuration.steps),
    deadlockAllowed: (s) => {
        if (flowIsCompleted(configuration.steps, s)) {
            // all steps completed
            return { success: true, value: null };
        } else {
            return { success: false, error: `The program has terminated or deadlocked. It was expected to ${flowDescribeNextStep(configuration.steps, s)}.` };
        }
    },
    cycleAllowed: (s, move) => configuration?.allowCycleOnMoves ? configuration.allowCycleOnMoves(move) : { success: false, error: "The program has entered an infinite loop." },
    moveSetAllowed: (s, moves) => {
        if (configuration.isMoveSetAllowed) {
            const moveSetAllowedResult = configuration.isMoveSetAllowed(moves);
            if (!moveSetAllowedResult.success) return moveSetAllowedResult; // this move set is banned
        }
        return { success: true, value: null };
    },
    acceptMove: (s, move) => {
        if (configuration.isLegalMove) {
            const legalMoveCheckResult = configuration.isLegalMove(move);
            if (!legalMoveCheckResult.success) return legalMoveCheckResult; // this move is globally banned
        }
        const res = flowAcceptMove(configuration.steps, s, move);
        if (res.success && configuration.restrictedMoves) for (const restriction of configuration.restrictedMoves) {
            // we made no progress, so we need to make sure no restricted move was executed
            const restrictionDetectorResult = restriction.detector(move);
            if (!restrictionDetectorResult.success) return restrictionDetectorResult;
            if (restrictionDetectorResult.value && JSON.stringify(res.value) === JSON.stringify(s)) return {
                success: false,
                error: `The program ${restriction.description} unexpectedly.`
            };
        }
        return res;
    },
    validationStateDescriber: (s) => flowGetValidationChecklistState(configuration.steps, s),
    endSearchAllowed: configuration.endSearchAllowed
});
