import { Move, parseExpression } from "@pseuco/colored-petri-nets";
import { ValueType } from "@pseuco/colored-petri-nets/dist/types";
import { getParticipatingAgents } from "./utils";
import { PartValidationResult } from "./validation";

/**
 * Validates a set of moves to determine whether they are allowed to occur together.
 */
export type MoveSetValidator = (moves: Move[]) => PartValidationResult<null>;


export const moveValidatorDisallowDataRaces = (): MoveSetValidator => (moves) => {
    const relevantMoves: {
        agent: number;
        variable: string;
        write: boolean;
        move: Move;
        path: ValueType[];
    }[] = [];

    // first, we aggregate the relevant data for each move
    for (const move of moves) {
        const type = move.transition.extensions.plusplus?.type as string;
        if (["read-global", "write-global", "update-global"].includes(type)) {
            const write = ["write-global", "update-global"].includes(type);

            const participatingAgents = getParticipatingAgents(move);
            if (participatingAgents.length !== 1) throw new Error("Found a global variable transition that is not executed by exactly one agent.");
            const agent = participatingAgents[0];

            const variable = move.transition.extensions.plusplus?.variable as string;

            let path: ValueType[] = [];
            try {
                const pathExpression = parseExpression(move.transition.extensions.plusplus?.path);
                const res = pathExpression(move.environment);
                if (res.length !== 1) throw new Error("Path expression did not produce exactly one result.");
                if (!(res[0] instanceof Array)) throw new Error("Path expression did not yield an array.");
                path = res[0];
            } catch (e) {
                throw new Error(`Error evaluation path of global variable transition: ${e.message}`);
            }
            
            relevantMoves.push({ agent, variable, write, move, path });
        }
    }

    // now, check for data races
    for (let i = 0; i < relevantMoves.length; i++) {
        for (let j = i + 1; j < relevantMoves.length; j++) {
            const move1 = relevantMoves[i];
            const move2 = relevantMoves[j];

            if (move1.agent === move2.agent) continue; // having two options for the same agent if OK
            if (move1.variable !== move2.variable) continue; // accessing different variables is OK
            if (!move1.write && !move2.write) continue; // reading concurrently is OK

            const pathPrefixLength = Math.min(move1.path.length, move2.path.length);
            if (JSON.stringify(move1.path.slice(0, pathPrefixLength)) !== JSON.stringify(move2.path.slice(0, pathPrefixLength))) continue; // accessing disjoint paths is OK

            // data race detected
            const variable = move1.variable;
            const pseuCoVariableNameMatch = /^pseuCo_(.*)$/.exec(variable);

            const humanReadableVariableName = pseuCoVariableNameMatch ? `pseuCo variable ${pseuCoVariableNameMatch[1]}` : variable === "arrays" ? `the same array` : variable === "structs" ? `the same struct or monitor` : `runtime variable ${variable}`;

            const readWriteSummary = move1.write && move2.write ? `Agents ${move1.agent} and ${move2.agent} both tried to perform a write access to ${humanReadableVariableName}.` : `Agent ${move1.write ? move1.agent : move2.agent} tried to write to ${humanReadableVariableName} while agent ${move1.write ? move2.agent : move1.agent} tried to read it.`;
            const pathSummary = pathPrefixLength > 0 ? ` (The access happened at path ${JSON.stringify(move1.path.slice(0, pathPrefixLength))}.)` : "";

            return {
                success: false,
                error: `The program has a data race. ${readWriteSummary}${pathSummary}`
            };
        }
    }

    return { success: true, value: null };
};
