import { Move, parseExpression, Place } from "@pseuco/colored-petri-nets";
import { ValueType } from "@pseuco/colored-petri-nets/dist/types";
import { simpleToStableJSON } from "@pseuco/colored-petri-nets/dist/utils";
import { get } from "../../../pseuco-shared-components/util/object";
import { getParticipatingAgents } from "./utils";
import { PartValidationResult } from "./validation";


const OK: { success: true, value: true } = { success: true, value: true };
const PASS: { success: true, value: false } = { success: true, value: false };


/**
 * Validates a single step of a [[LinearReachabilityGraphValidatorConfiguration]].
 * Inspects a move and returns a positive [[ValidationResult]] with `true` to indicate the move has completed a step or with `false` to indicate the step has not been performed yet.
 * A negative [[ValidationResult]] indicates an illegal action has been performed and validation should be failed immediately.
 */
export type MoveValidator = (move: Move) => PartValidationResult<boolean>;


/**
 * Turns a [[MoveValidator]] into one that does not output any value, e.g. allowing move validators to be used for `deadlockAllowed` and `cycleAllowed` of a [[ReachabilityGraphValidator]].
 * @param validator The validator to convert.
 */
export const dropMoveValidatorResultValue = (validator: MoveValidator) => (move: Move): PartValidationResult<null> => {
    const value = validator(move);
    if (value.success) return { success: true, value: null };
    else return value;
};


/**
 * Logical ∧ operation for validators.
 * @param validators An array of validators.
 */
export const moveValidatorAnd = (...validators: MoveValidator[]): MoveValidator => (move) => {
    let value = true; // whether all validators returned true so far

    for (const validator of validators) {
        const res = validator(move);
        if (!res.success) return res;
        if (!res.value) value = false;
    }

    return { success: true, value };
};


/**
 * Logical ∨ operation for validators.
 * @param validators An array of validators.
 */
export const moveValidatorOr = (...validators: MoveValidator[]): MoveValidator => (move) => {
    for (const validator of validators) {
        const res = validator(move);
        if (!res.success) return res;
        if (res.value) return OK;
    }

    return PASS;
};


/**
 * Negates a validator: Success will turn into rejection and vice-versa.
 * Neutral results stay neutral.
 * @param validator The validator to negate.
 * @param message The error message to return when the inner validator gave a positive return.
 */
export const moveValidatorNot = (validator: MoveValidator, message: string): MoveValidator => (move) => {
    const value = validator(move);
    if (value.success) {
        if (value.value) {
            return { success: false, error: message };
        } else {
            return PASS;
        }
    } else {
        return OK;
    }
};


/**
 * Combines a validator that detects a certain scenario with a requirement that will fail validation if it is missed.
 * Useful for scenarios like “after some computation, send a value, but the first value you send must be 42”.
 * @param detect The main validator. Its result is passed out unfiltered, except if it is `true`.
 * @param requirement This validator is run if `detect` returned `true`. Errors are passed through, and `false` returns an error.
 * @param message The message to return if `detect` was `true` but `require` was `false`.
 */
export const moveValidatorDetectAndRequire = (detect: MoveValidator, requirement: MoveValidator, message: string): MoveValidator => (move) => {
    const detectResult = detect(move);
    if (!detectResult.success) return detectResult;
    if (detectResult.value) {
        const requireResult = requirement(move);
        if (!requireResult.success) return requireResult;
        if (requireResult.value) {
            return OK;
        } else {
            return { success: false, error: message };
        }
    } else {
        return detectResult;
    }
};


/**
 * Checks whether a move prints an expected string.
 * @param allowedValues Used to check the printed strings.
 * @param allowOtherPrints If `true`, any non-matching print is allowed (but does not complete the step), otherwise, any other value is an error.
 */
export const moveValidatorPrintLn = (allowedValues: RegExp, allowOtherPrints: boolean): MoveValidator => (move) => {
    const printlnString = move.transition.extensions.println;
    if (printlnString) {
        const printlnExpression = parseExpression(printlnString);
        const printlnValues = printlnExpression(move.environment);
        let foundMatchingPrint = false;
        for (const printlnValue of printlnValues) {
            const match = allowedValues.test(printlnValue.toString());
            if (match) {
                foundMatchingPrint = true;
            } else {
                if (!allowOtherPrints) return { success: false, error: `The program unexpectedly printed: ${printlnValue}` };
            }
        }
        return { success: true, value: foundMatchingPrint };
    } else {
        return PASS;
    }
};


/**
 * Checks whether a move is performed by the expected agents.
 * Most moves have only one participating agent.
 * @param expectedAgents The ids of the agents that should participate in the move.
 */
export const moveValidatorParticipatingAgents = (expectedAgents: number[]): MoveValidator => (move) => {
    const participatingAgents = getParticipatingAgents(move);
    return { success: true, value: participatingAgents.every((a) => expectedAgents.includes(a)) && expectedAgents.every((a) => participatingAgents.includes(a)) };
};


/**
 * Checks whether a move is performed by a single agent at the expected recursion depth.
 */
export const moveValidatorRecursionDepth = (expectedDepth: number): MoveValidator => (move) => {
    const depths: number[] = [];

    Object.values(move.consumedTokens).forEach((ts) => ts.forEach((t) => {
        const rd = get(t.color, ["agent", "rd"], null);
        if (typeof rd === "number") depths.push(rd);
    }));

    if (depths.length === 1 && depths[0] === expectedDepth) {
        return OK;
    } else {
        return PASS;
    }
};


/**
 * Checks whether a message was passed on a handshaking channel.
 * @param checkValue Used to confirm the value was allowed.
 * @param allowOtherValues Whether other values are ignored or considered an error.
 */
export const moveValidatorHandshaking = (checkValue: (value: ValueType) => boolean, allowOtherValues: boolean, agents?: { sender?: number, receiver?: number, allowOtherAgents: boolean }): MoveValidator => (move) => {
    if (move.transition.extensions.plusplus?.type === "sync-handshake") {
        const senderAgent = get(move.environment, ["senderAgent", "id"], Error("Internal inconsistency: Could not detect sending agent in sync-handshake transition."));
        const receiverAgent = get(move.environment, ["receiverAgent", "id"], Error("Internal inconsistency: Could not detect sending agent in sync-handshake transition."));

        if ((agents?.sender !== undefined && agents?.sender !== senderAgent) || (agents?.receiver !== undefined && agents?.receiver !== receiverAgent)) {
            if (agents.allowOtherAgents) {
                return PASS;
            } else {
                return { success: false, error: "A handshaking action was performed between the wrong agents." };
            }
        }

        const value = move.environment["val"];
        if (value === undefined || value === null) throw new Error("Internal inconsistency: Could not detect value in sync-handshake transition.");
        if (checkValue(value)) {
            return OK;
        } else {
            if (allowOtherValues) {
                return PASS;
            } else {
                return { success: false, error: `The value ${JSON.stringify(value)} was sent (on a handshaking channel) unexpectedly.` };
            }
        }
    } else {
        return PASS;
    }
};


/**
 * Checks whether a message was passed on an asynchronous channel.
 * @param action Which async channel action to look for.
 * @param checkValue Used to confirm the value was allowed.
 * @param allowOtherValues Whether other values are ignored or considered an error.
 */
export const moveValidatorAsyncChannel = (action: "send" | "receive", checkValue: (value: ValueType) => boolean, allowOtherValues: boolean): MoveValidator => (move) => {
    if (move.transition.extensions.plusplus?.type === `${action}-async`) {
        const value = move.environment[action === "send" ? "s" : "b"];
        if (value === undefined || value === null) throw new Error(`Internal inconsistency: Could not detect value in ${action}-async transition.`);
        if (checkValue(value)) {
            return OK;
        } else {
            if (allowOtherValues) {
                return PASS;
            } else {
                return { success: false, error: `The value ${JSON.stringify(value)} was ${action === "send" ? "sent" : "received"} (on an asynchronous channel) unexpectedly.` };
            }
        }
    } else {
        return PASS;
    }
};


/**
 * Ensures that the current message passing transition was from a select-case-like situation.
 * When counting, this validator ignores handshaking transitions.
 * For nets generated from pseuCo, this does not matter, as every send and receive expression generates one async-send and async-receive transition (which are counted) and one handshake transition (which are ignored).
 * 
 * @param mode Whether to look for an asynchronous channel (which means only one agent participated), or a handshaking channel – and in the latter case, whether the sending or receiving agent should be inspected.
 * @param minSend The minimum number of send transitions.
 * @param minReceive The minimum number of receive transitions.
 * @param minDefault The minimum number of default transitions.
 * @param allowOther Whether to allow message-passing activities of the specified type that fail to meet the thresholds.
 */
export const moveValidatorSelectCase = (mode: "async" | "handshake-send" | "handshake-receive", minSend: number, minReceive: number, minDefault: number, allowOther: boolean): MoveValidator => (move) => {
    let sourcePlace: Place;

    // ensure we actually have the expected type of message passing action, and use our knowledge of compiler interna to find the source place of the agent token
    if (mode === "async") {
        if (!["send-async", "receive-async"].includes(move.transition.extensions.plusplus?.type)) return PASS;

        // we know that the first flow is the one carrying the agent
        sourcePlace = move.transition.inFlows[0].source;
    } else {
        if (move.transition.extensions.plusplus?.type !== "sync-handshake") return PASS;

        // we know that the sending flow is first, the receiving flow is second
        sourcePlace = move.transition.inFlows[mode === "handshake-send" ? 0 : 1].source;
    }

    const countTransitionsByType = (type: string) => Object.values(move.net.transitions).filter((t) => t.inFlows.some((flow) => flow.source === sourcePlace)).filter((t) => t.extensions.plusplus?.type === type).length;

    if (countTransitionsByType("send-async") >= minSend && countTransitionsByType("receive-async") >= minReceive && countTransitionsByType("default") >= minDefault) {
        return OK;
    } else {
        if (allowOther) {
            return PASS;
        } else {
            return { success: false, error: "This message passing action does not originate from a select-case statement with the required number of alternatives." };
        }
    }
};


/**
 * Checks whether any message passing was performed.
 * Use [[moveValidatorHandshaking]] or [[moveValidatorAsyncChannel]] for more detailed checks.
 */
export const moveValidatorMessagePassing = (): MoveValidator => moveValidatorOr(moveValidatorHandshaking(() => true, true), moveValidatorAsyncChannel("send", () => true, true), moveValidatorAsyncChannel("receive", () => true, true));


/**
 * Checks whether an agent was started.
 * Use [[moveValidatorParticipatingAgents]] to check which agents participated.
 */
export const moveValidatorStartAgent = (): MoveValidator => (move) => {
    if (move.transition.extensions.plusplus?.type === "start-agent") {
        return OK;
    } else {
        return PASS;
    }
};


/**
 * Checks whether an agent was terminated.
 * Use [[moveValidatorParticipatingAgents]] to check which agent was terminated.
 */
export const moveValidatorTerminateAgent = (): MoveValidator => (move) => {
    if (move.transition.extensions.plusplus?.type === "terminate") {
        return OK;
    } else {
        return PASS;
    }
};


/**
 * Checks whether a procedure was called.
 * @param validateArguments This callback is given the pseuCo arguments passed in the procedure call and decides whether the call counts, is ignored, or raised an error.
 */
export const moveValidatorCallProcedure = (validateArguments?: (args: { [name: string]: ValueType }) => PartValidationResult<boolean>): MoveValidator => (move) => {
    if (move.transition.extensions.plusplus?.type === "call") {
        if (validateArguments) {
            const outPlaces = Object.keys(move.producedTokens);
            if (outPlaces.length !== 2) throw new Error(`A call transition had an unexpected number of outgoing flows (${outPlaces.length}).`);
            const recursionDepths = outPlaces.map((key) => {
                const tokens = move.producedTokens[key];
                if (tokens.length !== 1) throw new Error(`A call transition had an outgoing flow that produced an unexpected number of tokens (${tokens.length})`);
                const rd = get(tokens[0], ["agent", "rd"], null);
                return typeof rd === "number" ? rd : -1;
            });
            const calleePlaceKey = recursionDepths[0] > recursionDepths[1] ? outPlaces[0] : outPlaces[1];
            const args = get(move.producedTokens[calleePlaceKey][0].color, ["vars", "vars"], null);
            if (typeof args !== "object" || args === null) throw new Error("Failed to identify arguments of procedure call.");

            return validateArguments(args as { [name: string]: ValueType });
        }

        return OK;
    } else {
        return PASS;
    }
};


/**
 * Checks whether a mutex is being initialized.
 * Note that this mutex may correspond to a pseuCo lock or be a managed monitor lock.
 */
export const moveValidatorInitializeMutex = (expectedId?: number, allowOther = false): MoveValidator => (move) => {
    if (move.transition.extensions.plusplus?.type === "init-mutex") {
        const id = get(move.environment, ["id"], new Error("Failed to identify mutex id."));

        if (expectedId !== undefined && id !== expectedId) {
            if (allowOther) return PASS;
            else return { success: false, error: "A lock was initialized unexpectedly." };
        }

        return OK;
    } else {
        return PASS;
    }
};


/**
 * Checks whether a mutex is locked or unlocked.
 * Note that this mutex may correspond to a pseuCo lock or be a managed monitor lock.
 */
export const moveValidatorLockUnlockMutex = (action: "lock" | "unlock", expectedId?: number, allowOther = false): MoveValidator => (move) => {
    if (move.transition.extensions.plusplus?.type === `${action}-mutex`) {
        const id = get(move.environment, ["mid"], new Error("Failed to identify mutex id."));

        if (expectedId !== undefined && id !== expectedId) {
            if (allowOther) return PASS;
            else return { success: false, error: `A lock was ${action}ed unexpectedly.` };
        }

        return OK;
    } else {
        return PASS;
    }
};


const makeMoveValidatorForPseuCoGlobalAccess = (action: "read" | "write") => (expectedName: string | null, expectedValue: ValueType | null, allowOther: boolean): MoveValidator => (move) => {
    const type = `${action}-global`;

    if (move.transition.extensions.plusplus?.type === type) {
        const variable = move.transition.extensions.plusplus?.variable;
        
        const pseuCoVariableMatch = /^pseuCo_(.*)$/.exec(variable);
        if (!pseuCoVariableMatch) return PASS;
        const pseuCoVariable = pseuCoVariableMatch[1];
    
        if (expectedName !== null) {
            if (pseuCoVariable !== expectedName) {
                if (allowOther) {
                    return PASS;
                } else {
                    return { success: false, error: `The program accessed global pseuCo variable ${pseuCoVariable} unexpectedly.` };
                }
            }
        }

        if (expectedValue !== null) {
            let foundValue = false;
            for (const placeKey in move.producedTokens) {
                const place = move.net.places[placeKey];
                if (place.extensions.plusplus?.type === "global-variable") {
                    const tokens = move.producedTokens[placeKey];
                    if (tokens.length !== 1) throw new Error("Global variable storage place has unexpected number of tokens.");
                    const value = tokens[0].color;

                    if (simpleToStableJSON(value) !== simpleToStableJSON(expectedValue)) {
                        if (allowOther) {
                            return PASS;
                        } else {
                            return { success: false, error: `A global variable was accessed with the unexpected value ${simpleToStableJSON(value)}.` };
                        }
                    }

                    foundValue = true;
                    break;
                }
            }
            if (!foundValue) throw new Error("Failed to find value of global variable access.");
        }

        return OK;
    } else {
        return PASS;
    }
};

export const moveValidatorReadPseuCoGlobal = makeMoveValidatorForPseuCoGlobalAccess("read");

export const moveValidatorWritePseuCoGlobal = makeMoveValidatorForPseuCoGlobalAccess("write");


/**
 * Checks for accesses to struct fields.
 * Note that monitors do counts as structs, here.
 */
const makeMoveValidatorForPseuCoStructFieldAccess = (action: "read" | "write") => (expectedId: number | null, expectedField: string | number | null, expectedValue: ValueType | null, allowOther: boolean): MoveValidator => (move) => {
    const type = `${action}-global`;

    if (move.transition.extensions.plusplus?.type === type) {
        const variable = move.transition.extensions.plusplus?.variable;
        
        if (variable !== `structs`) return PASS; // access to other global net variable, not relevant for us
    
        let path;
        try {
            path = parseExpression(move.transition.extensions.plusplus?.path)(move.environment)[0];
        } catch (e) {
            throw new Error("Internal inconsistency: Could not determine path for struct field access operation.");
        }

        if (!(path instanceof Array)) throw new Error("Internal inconsistency: Could not determine path for struct field access operation (not an array).");
        if (path.length < 3) return PASS; // this is not a simple read or write, but initialization or handling of implicit locks (for monitors)

        const [id,, field] = path;

        if (typeof id !== "number" || typeof field !== "string") throw new Error("Internal inconsistency: Unexpected type of id or field data in struct access operation.");

        if (expectedId !== null) {
            if (id !== expectedId) {
                if (allowOther) {
                    return PASS;
                } else {
                    return { success: false, error: `An access to struct ${id} was made unexpectedly.` };
                }
            }
        }

        if (expectedField !== null) {
            if (field !== expectedField) {
                if (allowOther) {
                    return PASS;
                } else {
                    return { success: false, error: `An access to struct field ${field} was made unexpectedly.` };
                }
            }
        }

        if (expectedValue !== null) {
            const structsTokens = move.marking?.tokens["global-structs"];
            if (!structsTokens || structsTokens.length !== 1) throw new Error("Internal inconsistency: Failed to find structs token.");
            const value = get(structsTokens[0].color, [id, "vars", field], Error("Internal inconsistency: Failed to extract struct field value.")) as ValueType;

            if (simpleToStableJSON(value) !== simpleToStableJSON(expectedValue)) {
                if (allowOther) {
                    return PASS;
                } else {
                    return { success: false, error: `A struct field was accessed with the unexpected value ${simpleToStableJSON(value)}.` };
                }
            }
        }

        return OK;
    } else {
        return PASS;
    }
};

export const moveValidatorReadPseuCoStruct = makeMoveValidatorForPseuCoStructFieldAccess("read");

export const moveValidatorWritePseuCoStruct = makeMoveValidatorForPseuCoStructFieldAccess("write");
