import { ReactElement, useContext, useState } from 'react';
import { parseAndCheckPseuCoFile } from '../../../pseuco-shared-components/parse/parseSupport';
import { InteractiveComponent } from '../../../hybrid-documents-reader-core/interactive/InteractiveElement';

import "./pseuCoProgramming.scss";
import programmingExerciseBuilder, { ValidationErrorVisualizer } from '../ProgrammingExercise';
import { PseuCoValidationError } from './validation';
import React from 'react';
import { Button, Modal } from 'react-bootstrap';
import { openFileInConcurrentProgrammingWeb } from '../../../pseuco-shared-components/util/pseuCoUrlFileEncoder';
import { CurrentThemeContext } from '../../../hybrid-documents-reader-core/ThemeApplicator';
import { TextFileData } from '../../../pseuco-shared-components/fileData';
import { PseuCoDebugger } from '../../../pseuco-shared-components/ui/debugger/PseuCoDebugger';

export type PseuCoProgrammingConfiguration = {
    allowedDialects: {
        mp: boolean;
        sm: boolean;
    }
    exerciseDescription: ReactElement,
    getWorker: () => Worker,
    template?: string,
    testValidation?: boolean,
    overrideSuccessMessage?: ReactElement
}

const PseuCoValidationErrorVisualization: ValidationErrorVisualizer<PseuCoValidationError> = ({ validationError, data }) => {
    const theme = useContext(CurrentThemeContext);

    const [debuggerModalData, setDebuggerModalData] = useState<null | {
        show: boolean,
        data: TextFileData,
        validationError: PseuCoValidationError,
        generation: number
    }>(null);
    
    const debuggerModal = debuggerModalData === null ? null : <Modal className="pseuco-debugger-modal" show={debuggerModalData.show} onHide={() => setDebuggerModalData({ ...debuggerModalData, show: false })}>
        <Modal.Header closeButton>
            <Modal.Title>pseuCo Debugger: Failing Trace</Modal.Title>
        </Modal.Header>
        <Modal.Body>
            { debuggerModalData.validationError.type === "BAD_TRACE" ? <PseuCoDebugger key={debuggerModalData.generation} data={debuggerModalData.data} net={debuggerModalData.validationError.net} mode={{
                type: "FIXED_TRACE",
                trace: debuggerModalData.validationError.trace.map((t, i) => ({
                    marking: t.marking,
                    move: t.move,
                    validationChecklistState: t.validationChecklistState,
                    error: debuggerModalData.validationError.type !== "BAD_TRACE" || i < debuggerModalData.validationError.trace.length - 1 ? null : debuggerModalData.validationError.message
                }))
            }} theme={theme} /> : null }
        </Modal.Body>
    </Modal>;

    return <>
        { debuggerModal }
        { validationError.type === "GENERIC" ? <p>There was a problem validating your program.<br />{ validationError.message }</p> : validationError.type === "SYNTAX_RESTRICTION_FAILED" ? <p>Your program does not satisfy a syntactic restriction posed by this exercise.<br />{ validationError.message }</p> : validationError.type === "SEARCH_ENDED_ILLEGALLY" ? <>
            <p>Your program failed to satisfy a requirement of this exercise.<br />{ validationError.message }</p>
            <Button className="debug-option-button" variant="secondary" onClick={() => openFileInConcurrentProgrammingWeb({ type: "pseuco", content: data.core ?? "" })}>Open in pseuCo.com</Button>
        </> : validationError.type === "BAD_TRACE" ? <>
            <p>During testing, we found an execution where your program does not work as expected.<br />{ validationError.message }</p>
            <Button className="debug-option-button" variant="secondary" onClick={() => setDebuggerModalData({
                show: true,
                data,
                validationError,
                generation: (debuggerModalData?.generation ?? 0) + 1
            })}>Show in Debugger</Button>
            <Button className="debug-option-button" variant="secondary" onClick={() => openFileInConcurrentProgrammingWeb({ type: "pseuco", content: data.core ?? "" })}>Open in pseuCo.com</Button>
        </> : <p>Unknown error.</p> }
    </>;
};

const pseuCoProgrammingBuilder = (configuration: PseuCoProgrammingConfiguration): InteractiveComponent => {
    const PseuCoProgramming: InteractiveComponent = programmingExerciseBuilder<PseuCoValidationError>({
        exerciseDescription: configuration.exerciseDescription,
        getWorker: configuration.getWorker,
        template: configuration.template,
        testValidation: configuration.testValidation,
        overrideSuccessMessage: configuration.overrideSuccessMessage,
        validateSyntax: (file) => {
            const { issues } = parseAndCheckPseuCoFile(file, configuration.allowedDialects);
            return issues;
        },
        ValidationErrorVisualization: PseuCoValidationErrorVisualization
    });

    return PseuCoProgramming;
};

export default pseuCoProgrammingBuilder;
