import { Program, Node, IdentifierExpression, PostfixExpression, AssignDestination, VariableDeclarations, ProcedureDeclaration, Struct, MainAgent, Monitor } from "@pseuco/lang";
import { PartValidationResult } from "./validation";

export type SyntaxValidator = (ast: Program) => PartValidationResult<null>;

/**
 * Limits how often a variable may be used.
 * @param variableName The name of the variable to limit.
 * @param maxCounts Maximum amount of usage in different contexts.
 * @returns 
 */
export const syntaxValidatorRestrictVariableUsage = (variableName: string, maxCounts: { identifier: number, postfix: number, assign: number }): SyntaxValidator => (ast) => {
    const counts: typeof maxCounts = {
        identifier: 0,
        postfix: 0,
        assign: 0
    };
    const contextDescriptions: { [context in keyof typeof counts]: string } = {
        identifier: "as an identifier",
        postfix: "within a postfix expression",
        assign: "on the left-hand-side of an assignment"
    };

    const processNode = (context: { inPostfix: boolean }) => (node: Node) => {
        if (node instanceof IdentifierExpression) {
            if (node.identifier === variableName) {
                counts.identifier++;
            }
        } else {
            if (node instanceof PostfixExpression) {
                processNode({ inPostfix: true })(node.assignDestination);
            } else if (node instanceof AssignDestination) {
                if (node.identifier === variableName) {
                    if (context.inPostfix) counts.postfix++;
                    else counts.assign++;
                }
            } else {
                // simply recurse into all children
                node.childNodes.forEach(processNode(context));
            }
        }
    };

    processNode({ inPostfix: false })(ast);

    const pluralize = (n: number, word: string) => n === 1 ? `${n} ${word}` : `${n} ${word}s`;

    let c: keyof typeof counts;
    for (c in counts) if (counts[c] > maxCounts[c]) {
        const restriction = maxCounts[c] > 0 ? `only be used ${pluralize(maxCounts[c], "time")}` : "not be used";
        const contextDescription = contextDescriptions[c];
        return { success: false, error: `Variable ${variableName} may ${restriction}${contextDescription.length ? " " + contextDescription : ""} (found ${pluralize(counts[c], "occurrence")}).` };
    }

    return { success: true, value: null };
};

/**
 * Prevents the use of any global variables.
 * Note that this still allows other kinds of shared memory (and blocks read-only global channel variables despite them being allowed in pseuCo-MP).
 */
export const syntaxValidatorDisallowGlobalVariables = (): SyntaxValidator => (ast) => {
    const globalVariables: string[] = [];

    const processNode = (node: Node) => {
        if (node instanceof VariableDeclarations) {
            node.getDeclarators().forEach((declarator) => globalVariables.push(declarator.name));
        } else if (node instanceof ProcedureDeclaration || node instanceof Struct || node instanceof Monitor || node instanceof MainAgent) {
            return;
        } else {
            throw new Error("Internal inconsistency: Unexpected top-level syntax element.");
        }
    };
    ast.childNodes.map(processNode);

    if (globalVariables.length <= 0) {
        return { success: true, value: null };
    } else {
        return {
            success: false,
            error: `Your program contains global variables: ${globalVariables.join(", ")}`
        };
    }
};


/**
 * Checks whether certain nodes are only used where they are allowed.
 * @param isRestrictedNode These node may not appear in the AST, except below any `isAllowingNode` node.
 * @param isAllowingNode Below these nodes, nothing is checked anymore.
 * @param error The user-facing error message returned if a violation is detected.
 */
// ↓ We truly mean Function here. Note that we do not call them, they are only on the right-hand side of instanceof checks.
// eslint-disable-next-line @typescript-eslint/ban-types
export const syntaxValidatorAllowNodeOnlyWithinNode = (isRestrictedNode: (node: Node) => boolean, isAllowingNode: (node: Node) => boolean, error: string): SyntaxValidator => (ast) => {
    let foundViolation = false;

    const processNode = (node: Node) => {
        if (isRestrictedNode(node)) {
            foundViolation = true;
        } else if (isAllowingNode(node)) {
            return; // nothing to do below those
        } else {
            node.childNodes.map(processNode);
        }
    };
    processNode(ast);

    if (foundViolation) {
        return { success: false, error };
    } else {
        return { success: true, value: null };
    }
};


/**
 * Fully bans certain nodes from occurring in the AST.
 * @param isBannedNode Determines whether a node is banned.
 * @param error The error message to return if a banned node is used.
 */
export const syntaxValidatorBanNode = (isBannedNode: (node: Node) => boolean, error: string): SyntaxValidator => syntaxValidatorAllowNodeOnlyWithinNode(isBannedNode, () => false, error);
