import { coloredPetriNetsPlusPlusCompiler, Marking, Move, PetriNet } from "@pseuco/colored-petri-nets";
import { MarkingObject, MoveObject, PetriNetObject } from "@pseuco/colored-petri-nets/dist/coloredPetriNets";
import { Compiler } from "@pseuco/pseuco-cpn-compiler";
import { TextFileData } from "../../../pseuco-shared-components/fileData";
import { parseAndCheckPseuCoFile, typeCheckPseuCoFile } from "../../../pseuco-shared-components/parse/parseSupport";
import { ValidationChecklistState } from "../../../pseuco-shared-components/ui/debugger/PseuCoDebugger";
import { SyntaxValidator } from "./syntaxValidation";
import { pickRandom } from "./utils";
import { ValidationResult } from "../ProgrammingExercise";

export type PartValidationError = string;

/**
 * A success flag and either a value of a generic type (for success) or an error message.
 */
export type PartValidationResult<T> = { success: true, value: T } | { success: false, error: PartValidationError }

export type PseuCoValidationTracePublicEntry = {
    marking: MarkingObject;
    validationChecklistState: ValidationChecklistState | null;
    move: MoveObject | null;
}

export type PseuCoValidationTraceEntry<PseuCoValidatorState> = PseuCoValidationTracePublicEntry & {
    validatorState: PseuCoValidatorState;
}

export type PseuCoValidationError = {
    type: "GENERIC",
    message: string
} | {
    type: "SYNTAX_RESTRICTION_FAILED",
    message: string
} | {
    type: "SEARCH_ENDED_ILLEGALLY",
    message: string
} | {
    type: "BAD_TRACE",
    reason: "BAD_MOVE" | "BAD_MOVE_SET" | "BAD_CYCLE" | "BAD_DEADLOCK";
    message: string;
    net: PetriNetObject;
    trace: PseuCoValidationTracePublicEntry[];
}

/**
 * An abstract validator for a reachability graph.
 * It accepts a path through the reachability graph if from the initial state, successively applying `acceptMove` never returns an error and after reaching the terminal marking, `deadlockAllowed` indicates the target state is allowed to deadlock.
 * If a cycle is detected, `cycleAllowed` is called on the state where the cycle was closed (but not on all states of the cycle!) to validate whether a cycle is allowed.
 * In addition, for every step, `moveSetAllowed` gets to inspect the whole set of enabled moves and can reject it.
 * Finally, at the end, `endSearchAllowed` is asked to confirm the search may and. This allows checking whether the whole search fulfilled a property that has been tracked externally (e.g. whether a certain move existed anywhere in the graph).
 */
export type ReachabilityGraphValidator<PseuCoValidatorState> = {
    initialState: PseuCoValidatorState,
    deadlockAllowed: (state: PseuCoValidatorState) => PartValidationResult<null>;
    cycleAllowed: (state: PseuCoValidatorState, move: Move) => PartValidationResult<null>;
    moveSetAllowed: (state: PseuCoValidatorState, moves: Move[]) => PartValidationResult<null>;
    endSearchAllowed?: () => PartValidationResult<null>;
    acceptMove: (state: PseuCoValidatorState, move: Move) => PartValidationResult<PseuCoValidatorState>;
    validationStateDescriber?: (s: PseuCoValidatorState) => ValidationChecklistState;
}


export const validateReachabilityGraph = <PseuCoValidatorState>(net: PetriNet, validator: ReachabilityGraphValidator<PseuCoValidatorState>, traceCount = -1): ValidationResult<PseuCoValidationError> => {
    type SearchState = {
        validatorState: PseuCoValidatorState,
        marking: Marking
    }

    /**
     * Turns a [[SearchState]] into a unique string for duplicate checking
     */
    const stringifySearchState = (searchState: SearchState): string => JSON.stringify({ vs: searchState.validatorState, m: searchState.marking.toStableJSON() });

    const initialState = { validatorState: validator.initialState, marking: net.initialMarking };
    let frontier: SearchState[] = [initialState];
    const seen: Map<string, SearchState | null> = new Map(); // In spirit, this has type Set<SearchState>, but because we only need this to check for duplicates (and never to actually look into it), we store stringified versions instead. And because we need to be able to identify predecessor states for cycle checking, we map to them.

    /**
     * Collects an error trace to return as part of a [[ValidationResult]].
     * @param searchState The state the error originates in.
     * @param move The move the error originated in.
     */
    const collectTrace = (searchState: SearchState, move: Move | null): PseuCoValidationTraceEntry<PseuCoValidatorState>[] => {
        const res: PseuCoValidationTraceEntry<PseuCoValidatorState>[] = [];

        res.push({
            marking: searchState.marking.toObject(),
            validatorState: searchState.validatorState,
            validationChecklistState: validator.validationStateDescriber ? validator.validationStateDescriber(searchState.validatorState) : null,
            move: move === null ? null : move.toObject()
        });

        while (true) {
            const nextSearchState = seen.get(stringifySearchState(searchState));
            if (!nextSearchState) return res;
            
            const moves = nextSearchState.marking.enabledMoves();
            const targetMarking = searchState.marking.toStableJSON();
            const nextMove = moves.find((m) => m.marking?.toStableJSON() === targetMarking);
            if (!nextMove) throw new Error("Internal inconsistency: Failed to find move from previous marking.");
    
            res.unshift({
                marking: nextSearchState.marking.toObject(),
                validatorState: nextSearchState.validatorState,
                validationChecklistState: validator.validationStateDescriber ? validator.validationStateDescriber(nextSearchState.validatorState) : null,
                move: nextMove.toObject()
            });

            searchState = nextSearchState;
            move = nextMove;
        }
    };

    let attempts = traceCount;
    do {
        frontier = [initialState];
        seen.clear();
        seen.set(stringifySearchState(initialState), null);
    
        while (true) {
            const currentJob = frontier.pop();
            if (currentJob) {
                const moves = currentJob.marking.enabledMoves();
    
                const movesValidatorResult = validator.moveSetAllowed(currentJob.validatorState, moves);
                if (!movesValidatorResult.success) return {
                    success: false,
                    error: {
                        type: "BAD_TRACE",
                        reason: "BAD_MOVE_SET",
                        message: movesValidatorResult.error,
                        net: net.toObject(),
                        trace: collectTrace(currentJob, null)
                    }
                };
    
                if (moves.length > 0) {
                    for (const move of traceCount >= 0 ? [pickRandom(moves)] : moves) {
                        if (!move.marking) throw new Error("Internal inconsistency: move has no marking despite ignoreInFlowTokenCounts = false");
    
                        const validatorResult = validator.acceptMove(currentJob.validatorState, move);
                        if (!validatorResult.success) return {
                            success: false,
                            error: {
                                type: "BAD_TRACE",
                                reason: "BAD_MOVE",
                                message: validatorResult.error,
                                net: net.toObject(),
                                trace: collectTrace(currentJob, move)
                            }
                        };
    
                        // move is allowed, do we know the target configuration?
                        const newSearchState: SearchState = { validatorState: validatorResult.value, marking: move.marking };
                        const newSearchStateString = stringifySearchState(newSearchState);
                        if (seen.has(newSearchStateString)) {
                            // we've seen this configuration before – did we simply re-discover it or is it a cycle?
                            let currentCycleCheckState: SearchState | null | undefined = currentJob;
                            let currentCycleCheckStateString = stringifySearchState(currentCycleCheckState); // cached for performance

                            while (true) {
                                if (currentCycleCheckStateString === newSearchStateString) {
                                    // found a cycle
                                    const validatorResult = validator.cycleAllowed(newSearchState.validatorState, move);
                                    if (validatorResult.success) {
                                        // cycle was allowed, cancel cycle search
                                        break;
                                    } else {
                                        // illegal cycle, report
                                        return {
                                            success: false,
                                            error: {
                                                type: "BAD_TRACE",
                                                reason: "BAD_CYCLE",
                                                message: validatorResult.error,
                                                net: net.toObject(),
                                                trace: collectTrace(currentJob, move)
                                            }
                                        };
                                    }
                                }

                                currentCycleCheckState = seen.get(currentCycleCheckStateString);
                                if (!currentCycleCheckState) break; // we arrived at the initial state
                                
                                currentCycleCheckStateString = stringifySearchState(currentCycleCheckState);
                            }
                        } else {
                            // new configuration, add to queue
                            seen.set(newSearchStateString, currentJob);
                            frontier.unshift(newSearchState);
                        }
                    }
                } else {
                    // terminal state
                    const validatorResult = validator.deadlockAllowed(currentJob.validatorState);
                    if (!validatorResult.success) return {
                        success: false,
                        error: {
                            type: "BAD_TRACE",
                            reason: "BAD_DEADLOCK",
                            message: validatorResult.error,
                            net: net.toObject(),
                            trace: collectTrace(currentJob, null)
                        }
                    };
                }
            } else {
                attempts--;
                break;
            }
        }
    } while (attempts > 0);

    if (validator.endSearchAllowed) {
        const endSearchResult = validator.endSearchAllowed();
        if (endSearchResult.success) {
            return { success: true };
        } else  {
            return {
                success: false,
                error: {
                    type: "SEARCH_ENDED_ILLEGALLY",
                    message: endSearchResult.error
                }
            };
        }
    } else {
        return { success: true };
    }
};


/**
 * Expected to be called in a WebWorker. Registers the `onmessage` callback to receive programs and run the validator against it.
 * @param validator The [[ReachabilityGraphValidator]] to use in this worker.
 */
export const registerValidationWorkerCallback = <PseuCoValidatorState>(validator: ReachabilityGraphValidator<PseuCoValidatorState>, allowedDialects: { mp: boolean, sm: boolean }, syntaxChecks: SyntaxValidator[] = [], traceCount = -1): void => {
    const validateProgram = (file: TextFileData): ValidationResult<PseuCoValidationError> => {
        try {
            // parse
            const { issues, ast } = parseAndCheckPseuCoFile(file, allowedDialects);
            if (issues.some((i) => i.type === "error") || ast === null) return {
                success: false,
                error: {
                    type: "GENERIC",
                    message: "Validation failed because the program could not be compiled."
                }
            };

            // run syntax checks
            for (const check of syntaxChecks) {
                const res = check(ast);
                if (!res.success) return {
                    success: false,
                    error: {
                        type: "SYNTAX_RESTRICTION_FAILED",
                        message: res.error
                    }
                };
            }

            // compile
            const pseuCoCppnCompiler = new Compiler();
            const { staticSemantics, success: typeCheckSuccess } = typeCheckPseuCoFile(ast);
            if (!typeCheckSuccess) return {
                success: false,
                error: {
                    type: "GENERIC",
                    message: `The program unexpectedly failed to typecheck in the validation worker.`
                }
            };
            const cppn = pseuCoCppnCompiler.compile(ast, staticSemantics);
            const cpn = coloredPetriNetsPlusPlusCompiler.compile(cppn);

            // run semantic check
            return validateReachabilityGraph(cpn, validator, traceCount);
        } catch (e) {
            return {
                success: false,
                error: {
                    type: "GENERIC",
                    message: `An internal error occurred during validation: ${e.message}`
                }
            };
        }
    };

    self.onmessage = ({ data: file }: { data: TextFileData }) => {
        self.postMessage(validateProgram(file));
    };
};
