import "./vendingMachines.scss";
import { InteractiveComponent } from "../../../hybrid-documents-reader-core/interactive/InteractiveElement";
import React, { useState, ReactElement, Fragment, useContext, useEffect } from "react";

import { LTSViewData } from "../../../pseuco-shared-components/ui/editors/lts/LTSInteractiveView";

import vendingMachine from "./vending-machine.svg";
import { Button } from "react-bootstrap";
import LTSSVGFrame from "../../../pseuco-shared-components/ui/editors/lts/LTSSVGFrame";
import LTSSVGFragmentView, { LTSTransitionViewData } from "../../../pseuco-shared-components/ui/editors/lts/LTSSVGFragmentView";
import { shiftToTransform, DoubleSided } from "../DoubleSidedLTS";
import { UserStudyContext } from "../../study/UserStudy";

const VendingMachineBuilder = async (lts: DoubleSided<LTSViewData>, names: DoubleSided<string>, {
    checkTraces = false,
    checkTauRelevance = false
}: {
    checkTraces?: boolean;
    checkTauRelevance?: boolean;
}): Promise<InteractiveComponent> => {
    const VendingMachine: React.FC<{
        title: string;
        lts: LTSViewData;
        showLts: boolean;

        /**
         * Resetting this to a new object will reset the state of the simulation
         */
        simulationGenerationToken: {};
    }> = ({ lts, title, showLts, simulationGenerationToken }) => {
        const [state, setState] = useState<string>("Machine");

        useEffect(() => {
            setState("Machine");
        }, [simulationGenerationToken]);

        const enabledTransitions = lts.transitions.filter((t) => t.source === state);
    
        const getTransitionForAction = (action: string): LTSTransitionViewData | null => {
            const matches = enabledTransitions.filter((t) => t.label === action);
            if (matches.length === 0) return null;
            return matches[Math.floor(Math.random() * matches.length)];
        };
    
        const insertCoinTransition: LTSTransitionViewData | null = getTransitionForAction("coin?");
    
        const selectWaterTransition: LTSTransitionViewData | null = getTransitionForAction("selectWater?");
        const selectLemonadeTransition: LTSTransitionViewData | null = getTransitionForAction("selectLemonade?");
        const selectCoffeeTransition: LTSTransitionViewData | null = getTransitionForAction("selectCoffee?");
    
        const deliverWaterTransition: LTSTransitionViewData | null = getTransitionForAction("water!");
        const deliverLemonadeTransition: LTSTransitionViewData | null = getTransitionForAction("lemonade!");
        const deliverCoffeeTransition: LTSTransitionViewData | null = getTransitionForAction("coffee!");

        const timeTickTransition: LTSTransitionViewData | null = getTransitionForAction("τ");
    
        const statusMessage = insertCoinTransition ? "NO CREDIT" : selectWaterTransition || selectLemonadeTransition || selectCoffeeTransition ? "SELECT DRINK" : deliverWaterTransition || deliverLemonadeTransition || deliverCoffeeTransition ? "♥ THANK YOU ♥" : "SYSTEM ERROR";
    
        const takeTransition = (transition: LTSTransitionViewData): void => setState(transition.target);
    
        const [tickedOnceAlready, setTickedOnceAlready] = useState<boolean>(false);
        const getRandomDelay = (): number => {
            if (!tickedOnceAlready) return 1000;
            return 500 + Math.pow(Math.random(), 2) * 9500;
        };

        const [tickTimeout, setTickTimeout] = useState<ReturnType<typeof setTimeout> | null>(null);
        if (timeTickTransition && tickTimeout === null) {
            setTickTimeout(setTimeout(() => {
                takeTransition(timeTickTransition);
                setTickedOnceAlready(true);
            }, getRandomDelay()));
        }
        if (!timeTickTransition && tickTimeout !== null) {
            clearTimeout(tickTimeout);
            setTickTimeout(null);
        }

        const getHighlight = (): ReactElement => <circle className="highlight" r="30" style={{ transform: shiftToTransform(lts.states[state])}} />;
    
        return <div className="vending-machine">
            <div className="vending-main">
                <div className="title">{ title }</div>
                <img className="vending-image" src={vendingMachine} />
                { showLts ? <LTSSVGFrame darkMode={true} hideWeak={false} largeFont={false} moreSvgProps={{ viewBox: `85 -10 140 350` }}>
                    { getHighlight() }
                    <LTSSVGFragmentView states={lts.states} transitions={lts.transitions} showExpandNotice={false} />
                </LTSSVGFrame> : null }
                <div className="status-message">{ statusMessage }</div>
                { insertCoinTransition ? <Button className="insert-coin" onClick={(): void => takeTransition(insertCoinTransition)}>Insert Coin</Button> : null }
                { selectWaterTransition ? <Button className="select-water" onClick={(): void => takeTransition(selectWaterTransition)}>Water</Button> : null }
                { selectLemonadeTransition ? <Button className="select-lemonade" onClick={(): void => takeTransition(selectLemonadeTransition)}>Lemonade</Button> : null }
                { selectCoffeeTransition ? <Button className="select-coffee" onClick={(): void => takeTransition(selectCoffeeTransition)}>Coffee</Button> : null }
                { deliverWaterTransition ? <Button className="deliver-water" onClick={(): void => takeTransition(deliverWaterTransition)}>Take Water</Button> : null }
                { deliverLemonadeTransition ? <Button className="deliver-lemonade" onClick={(): void => takeTransition(deliverLemonadeTransition)}>Take Lemonade</Button> : null }
                { deliverCoffeeTransition ? <Button className="deliver-coffee" onClick={(): void => takeTransition(deliverCoffeeTransition)}>Take Coffee</Button> : null }
            </div>
        </div>;
    };
    
    const VendingMachines: InteractiveComponent = ({ state, setState, markAsSolved, id }) => {
        const userStudy = useContext(UserStudyContext);

        const [machineSelectionError, setMachineSelectionError] = useState<boolean>(false);
        const [equalityError, setEqualityError] = useState<boolean>(false);
    
        let machineSelected = false;
        let equalityConfirmed = false;
        if (state) {
            if (typeof state.machineSelected === "boolean") machineSelected = state.machineSelected;
            if (typeof state.equalityConfirmed === "boolean") equalityConfirmed = state.equalityConfirmed;
        }
    
        const saveState = (): void => {
            if (equalityConfirmed || (machineSelected && !checkTraces && !checkTauRelevance)) markAsSolved();
            setState({
                machineSelected,
                equalityConfirmed
            });
        };

        const stage = machineSelected ? (equalityConfirmed ? 2 : 1) : 0;
        useEffect(() => {
            userStudy.reportEvent({ type: "INTERACTIVE_PROGRESS_UPDATE", interactiveType: "interactive", interactiveId: id, stage, progress: 0 });
        }, [id, stage, userStudy]);
    
        const reset = (): void => setState(null);
    
        const [simulationGenerationToken, setSimulationGenerationToken] = useState<{}>({});

        const reportError = () => userStudy.reportEvent({ type: "INTERACTIVE_REPORT_ERROR", interactiveType: "interactive", interactiveId: id });

        return <div className="trace-equivalent-vending-machines">
            <p>In this exercise, we consider two different vending machines for drinks. Play a bit with both to understand how they behave. (If you wish to return both machines to their initial state, you can <a className="reset-simulation" onClick={() => setSimulationGenerationToken({})}>reset the simulation</a>.)</p>
            <div className="vending-area">
                <VendingMachine title={names.left} lts={lts.left} showLts={machineSelected} simulationGenerationToken={simulationGenerationToken} />
                <VendingMachine title={names.right} lts={lts.right} showLts={machineSelected} simulationGenerationToken={simulationGenerationToken} />
            </div>
            <p>
                After you have used both machines a few times, answer this question: Which machine works better?<br />
                <Button className="answer" onClick={(): void => {
                    setMachineSelectionError(false);
                    machineSelected = true;
                    saveState();
                } }>{ names.left }</Button>
                <Button disabled={machineSelected} className="answer" onClick={(): void => { reportError(); setMachineSelectionError(true); } }>{ names.right }</Button>
                <Button disabled={machineSelected} className="answer" onClick={(): void => { reportError(); setMachineSelectionError(true); } }>no difference</Button>
                { machineSelectionError ? <span className="think-again">Think again.</span> : null }
            </p>
            { machineSelected  ? <div>
                <p>The underlying LTS models of the machines are now shown above. Inspect them carefully. The current state of each machine is highlighted.</p>
                { checkTraces ? <Fragment>
                    <p>
                        Do both machines have different traces?<br />
                        <Button disabled={equalityConfirmed} className="answer" onClick={(): void => { reportError(); setEqualityError(true); } }>Yes, Holger's Drinks has more traces</Button>
                        <Button disabled={equalityConfirmed} className="answer" onClick={(): void => { reportError(); setEqualityError(true); } }>Yes, Felix' Refreshment Store has more traces</Button>
                        <Button disabled={equalityConfirmed} className="answer" onClick={(): void => { reportError(); setEqualityError(true); } }>Yes, the sets are incomparable</Button>
                        <Button className="answer" onClick={(): void => {
                            setEqualityError(false);
                            equalityConfirmed = true;
                            saveState();
                        } }>No, the traces are identical</Button>
                        { equalityError ? <span className="think-again">Think again.</span> : null }
                    </p>
                </Fragment> : null }

                { checkTauRelevance ? <Fragment>
                    <p>
                        If you merged source and target states of τ transitions, how would the underlying LTS compare?
                        Select the first correct answer.
                        <Button className="answer" onClick={(): void => {
                            setEqualityError(false);
                            equalityConfirmed = true;
                            saveState();
                        } }>they would be identical</Button>
                        <Button disabled={equalityConfirmed} className="answer" onClick={(): void => { reportError(); setEqualityError(true); } }>they would be bisimilar</Button>
                        <Button disabled={equalityConfirmed} className="answer" onClick={(): void => { reportError(); setEqualityError(true); } }>they would be trace equivalent</Button>
                        <Button disabled={equalityConfirmed} className="answer" onClick={(): void => { reportError(); setEqualityError(true); } }>neither</Button>
                        { equalityError ? <span className="think-again">Think again.</span> : null }
                    </p>
                </Fragment> : null }
    
                { equalityConfirmed ? <Fragment>
                    <p><b>That's it.</b>
                        { checkTraces ? <Fragment> To sum up: We've seen that even though LTS can have the same set of traces, they can have an observable difference.</Fragment> : null }
                        { checkTauRelevance ? <Fragment> To sum up: We've seen that even though LTS would be identical if we merged states along τ-transitions, they can have an observable difference.</Fragment> : null }
                    </p>
                    <div>Want to start over? <Button size="sm" variant="danger" className="reset-exercise" onClick={(): void => reset() }>Reset Exercise</Button></div>
                </Fragment> : null }
            </div> : null }
        </div>;
    };
    
    return VendingMachines;
};

export default VendingMachineBuilder;
